# -*- coding: utf-8 -*-
"""
Functions for cleaning and processing the AHBA microarray dataset
"""

import logging
from pkg_resources import resource_filename

import nibabel as nib
import numpy as np
import pandas as pd
from scipy.spatial.distance import cdist

from . import io, utils

lgr = logging.getLogger('abagen')

# AHBA structure IDs corresponding to different brain parts
ONTOLOGY = nib.volumeutils.Recoder(
    (('4008', 'cerebral cortex', 'cortex'),
     ('4275', 'cerebral nuclei', 'subcortex'),
     ('4391', 'diencephalon', 'subcortex'),
     ('9001', 'mesencephalon', 'subcortex'),
     ('4696', 'cerebellum', 'cerebellum'),
     ('9131', 'pons', 'brainstem'),
     ('9512', 'myelencephalon', 'brainstem'),
     ('9218', 'white matter', 'white matter'),
     ('9352', 'sulci & spaces', 'other'),
     ('4219', 'hippocampal formation', 'subcortex')),
    fields=('id', 'name', 'structure')
)


def update_mni_coords(annotation):
    """
    Replaces MNI coords in `annotation` with corrected coords from `alleninf`

    Parameters
    ----------
    annotation : str
        Annotation file from Allen Brain Institute. Optimally obtained by
        calling `abagen.fetch_microarray()` and accessing the `annotation`
        attribute on the resulting object

    Returns
    -------
    corrected : pandas.DataFrame
        Annotation data with corrected MNI coordinates

    References
    ----------
    Updated MNI coordinates taken from https://github.com/chrisfilo/alleninf,
    which is licensed under the BSD-3 (reproduced here):

    Copyright (c) 2018, Krzysztof Gorgolewski
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the names of its
      contributors may be used to endorse or promote products derived from
      this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
    INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
    CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
    POSSIBILITY OF SUCH DAMAGE.
    """

    coords = resource_filename('abagen', 'data/corrected_mni_coordinates.csv')
    coords = pd.read_csv(coords).rename(dict(corrected_mni_x='mni_x',
                                             corrected_mni_y='mni_y',
                                             corrected_mni_z='mni_z'),
                                        axis=1)
    coords = coords.set_index('well_id')

    annotation = io.read_annotation(annotation, copy=True)

    # basic check that all well_ids in annotation are present in coords
    # a future pandas update may cause this to raise a KeyError but we want
    # this to raise a KeyError NOW
    diff = np.setdiff1d(annotation['well_id'], coords.index)
    if len(diff) > 0:
        raise KeyError('Provided annotation file has well IDs that do not '
                       'exist in updated MNI coordinate file from `alleninf`. '
                       'Please check input annotation file and try again. '
                       'Unknown well IDs: {}'.format(diff))

    mni_coords = coords.loc[annotation.well_id]
    annotation[['mni_x', 'mni_y', 'mni_z']] = np.asarray(mni_coords)

    return annotation


def update_coords(annotation, corrected_mni=True, native_space=None):
    """
    Updates coordinates in `annotation`

    Parameters
    ----------
    annotation : str
        Annotation file from Allen Brain Institute. Optimally obtained by
        calling `abagen.fetch_microarray()` and accessing the `annotation`
        value for one of the returned donors
    corrected_mni : bool
        Whether to replace MNI coordinates in `annotation` with coordinates
        from `alleninf` (uses :func:`update_mni_coords`)
    native_space : niimg-like
        Whether to replace MNI coordinates in `annotation` with native-space
        coordinates; uses `affine` of provided image to convert coordinates

    Returns
    -------
    corrected : pandas.DataFrame
        Annotation data with updated coordinates
    """

    annotation = io.read_annotation(annotation, copy=True)

    if corrected_mni:
        annotation = update_mni_coords(annotation)

    if native_space is not None:
        try:
            native_space = nib.load(native_space)
        except TypeError:
            native_space = native_space
        affine = native_space.affine
        ijk = annotation[['mri_voxel_x', 'mri_voxel_y', 'mri_voxel_z']]
        annotation[['mni_x', 'mni_y', 'mni_z']] = utils.ijk_to_xyz(ijk, affine)

    return annotation


def _get_struct(structure_path):
    """
    Gets overarching "structure" of region defined by `structure_path`

    Structure here is defined as being one of {'cortex', 'subcortex',
    'cerebellum', 'brainstem', 'white matter', 'other'}.

    Parameters
    ----------
    structure_path : str
        As obtained from an ontology file from Allen Brain Institute. Optimally
        obtained by calling `abagen.fetch_microarray()`, loading the
        `ontology` attribute on the resulting object, and querying the
        `structure_id_path` column from the resulting dataframe

    Returns
    -------
    structure : str
        Structure, or None if unable to identify a corresponding structure
    """

    structures = set(structure_path.strip('/').split('/'))
    ids = sorted(set(ONTOLOGY.value_set('id')) & structures,
                 key=lambda x: structure_path.index(x))

    try:
        return ONTOLOGY.structure[ids[-1]]
    except IndexError:
        return


def drop_mismatch_samples(annotation, ontology):
    """
    Removes samples from `annotation` whose coordinates do not match `ontology`

    Checks MNI coordinates specified in `annotation` and matches them to L/R
    hemisphere designation in `ontology`; samples who do not match (e.g., L
    hemisphere designation in `ontology` but X coordinate < 0) are removed.

    Parameters
    ----------
    annotation : str
        Annotation file from Allen Brain Institute. Optimally obtained by
        calling `abagen.fetch_microarray()` and accessing the `annotation`
        attribute on the resulting object
    ontology : str
        Ontology file from Allen Brain Institute. Optimally obtained by
        calling `abagen.fetch_microarray()` and accessing the `ontology`
        attribute on the resulting object

    Returns
    -------
    annot : pandas.DataFrame
        Annotation data
    """

    # read in data files
    annotation = io.read_annotation(annotation)
    ontology = io.read_ontology(ontology).set_index('id')
    sid = np.asarray(annotation['structure_id'])

    # get hemisphere and structure path
    hemisphere = np.asarray(ontology.loc[sid, 'hemisphere'])
    structure = np.asarray(ontology.loc[sid, 'structure_id_path']
                                   .apply(_get_struct))

    # add hemisphere + brain "structure" designation to annotation data and
    # only keep samples with consistent hemisphere + MNI coordinate designation
    annot = annotation.assign(hemisphere=hemisphere, structure=structure) \
                      .query('(hemisphere == "L" & mni_x < 0) '
                             '| (hemisphere == "R" & mni_x > 0) '
                             '| (hemisphere.isna() & mni_x == 0)',
                             engine='python')

    return annot


def _mirror_ontology(annotation, ontology):
    """
    Assumes all hemisphere assignments of structures in `annotation` are wrong

    Parameters
    ----------
    annotation,ontology : str
        Filepath to {annotation,ontology} file

    Returns
    -------
    annotation : pandas.DataFrame
        Loaded annotation with updated structure information
    """

    HEMI_SWAP = dict(L='R', R='L')

    annotation = io.read_annotation(annotation, copy=True)
    ontology = io.read_ontology(ontology)

    # structure IDs are specific to structure + hemisphere, so we can use this
    # to grab the hemisphere designation for each sample and flip that L<->R
    hemi = ontology.set_index('id') \
                   .loc[annotation['structure_id'], 'hemisphere'] \
                   .replace(HEMI_SWAP)

    # structure acronyms are distinct to structure but not to hemisphere, so we
    # can use this to grab all variations of a given structure and, with the
    # flipped hemisphere designations we just generated, find the ID/name of
    # the relevant structure in the new hemisphere
    acr = ontology.set_index(['acronym', 'hemisphere']) \
                  .loc[zip(annotation['structure_acronym'], hemi)] \
                  .reset_index()

    # update the original annotation with the new values
    annotation['structure_acronym'] = np.asarray(acr['acronym'])
    annotation['structure_id'] = np.asarray(acr['id'])
    annotation['structure_name'] = np.asarray(acr['name'])

    return annotation


def mirror_samples(annotation, ontology):
    """
    Mirrors all tissue samples across hemispheres (L->R and R->L)

    Methodology follows [SA1]_ in that samples are mirrored bi-directionally
    rather than uni-directionally (R->L) as in [SA2]_.

    Parameters
    ----------
    annotation : str or pandas.DataFrame
        Filepath to annotation file from Allen Brain Institute (i.e., as
        obtained by calling :func:`abagen.fetch_microarray` and accessing the
        `annotation` attribute on the resulting object).
    ontology : str or pandas.DataFrame
        Filepath to ontology file from Allen Brain Institute (i.e., as
        obtained by calling :func:`abagen.fetch_microarray` and accessing the
        `ontology` attribute on the resulting object).

    Returns
    -------
    annotation : pandas.DataFrame
        Loaded input `annotation` data with all samples duplicated across
        hemispheres (where structure IDs and MNI coordinates are updated,
        accordingly)

    References
    ----------
    .. [SA1] Gryglewski, G., Seiger, R., James, G. M., Godbersen, G. M.,
       Komorowski, A., Unterholzner, J., ... & Kasper, S. (2018). Spatial
       analysis and high resolution mapping of the human whole-brain
       transcriptome for integrative analysis in neuroimaging. NeuroImage, 176,
       259-267.

    .. [SA2] Romero-Garcia, R., Whitaker, K. J., Váša, F., Seidlitz, J., Shinn,
       M., Fonagy, P., ... & Vértes, P. E. (2018). Structural covariance
       networks are coupled to expression of genes enriched in supragranular
       layers of the human cortex. NeuroImage, 171, 256-267.
    """

    annotation = io.read_annotation(annotation)
    ontology = io.read_ontology(ontology)

    # take all lh and rh samples and flip x-coordinate
    # also update ontology information (structure_id/acronym/name) as this is
    # used when dropping mismatched samples later in the workflow
    lh = _mirror_ontology(annotation[annotation['mni_x'] < 0], ontology)
    rh = _mirror_ontology(annotation[annotation['mni_x'] > 0], ontology)

    # update mni-x coordinate (simply flip the sign)
    for df in [lh, rh]:
        df['mni_x'] *= -1

    # grow microarray and pacall based on duplicated samples
    annotation = pd.concat([annotation, lh, rh])

    return annotation


def _assign_sample(sample, atlas, sample_info=None, atlas_info=None,
                   tolerance=2):
    """
    Determines which parcel `sample` belongs to in `atlas`

    Parameters
    ----------
    sample : (1, 3) array_like
        Coordinates (ijk) of microarray sample in `atlas` space
    atlas : niimg-like object
        ROI image, where each ROI should be identified with a unique
        integer ID
    sample_info : pandas.DataFrame
        A single row of an `annotation` file, corresponding to the given sample
    atlas_info : pandas.DataFrame,
        Dataframe containing information about the specified `atlas`. Must have
        _at least_ columns 'id', 'hemisphere', and 'structure' containing
        information mapping atlas IDs to hemisphere and broad structural class
        (i.e., "cortex", "subcortex", "cerebellum"). Default: None
    tolerance : int, optional
        Distance (in mm) that a sample must be from a parcel for it to be
        matched to that parcel. This is only considered if the sample is not
        directly within a parcel. Default: 2

    Returns
    -------
    label : int
        Parcel label of `sample`
    """

    # pull relevant info from atlas
    atlas_data = np.asarray(atlas.dataobj)

    # expand provided coordinates to include those w/i `tolerance` of `coords`
    # set a hard euclidean distance limit to account for different voxel sizes
    coords = utils.expand_roi(sample, dilation=tolerance, return_array=True)
    coords = coords[cdist(sample, coords).squeeze() < tolerance]

    # grab non-zero labels for expanded coordinates
    possible_labels = atlas_data[coords[:, 0], coords[:, 1], coords[:, 2]]
    nz_labels = possible_labels[possible_labels.nonzero()]
    labels, counts = np.unique(nz_labels, return_counts=True)

    # if atlas_info and sample_info are provided, drop potential labels who
    # don't match hemisphere or structural class defined in `sample_info`
    if atlas_info is not None and sample_info is not None:
        for old_label in labels:
            new_label = _check_label(old_label, sample_info, atlas_info)
            if old_label != new_label:
                nz_labels[nz_labels == old_label] = new_label
        labels, counts = np.unique(nz_labels[nz_labels.nonzero()],
                                   return_counts=True)

    # if there is still nothing in the vicinity, return 0
    if labels.size == 0:
        return 0
    # if there is only one ROI in the vicinity, use that
    elif labels.size == 1:
        return labels[0]

    # if more than one ROI in the vicinity, return the most frequent
    indmax, = np.where(counts == counts.max())
    if indmax.size == 1:
        return labels[indmax[0]]

    # if two or more parcels tied for neighboring frequency, use ROI
    # with closest centroid to `coords`
    centroids = utils.get_centroids(atlas, labels)
    return labels[utils.closest_centroid(sample, centroids)]


def _check_label(label, sample_info, atlas_info):
    """
    Checks that `label` defined by `sample_info` is coherent with `atlas_info`

    Parameters
    ----------
    label : int
        Tenative label for sample described by `sample_info`
    sample_info : pandas.DataFrame
        A single row of an `annotation` file, corresponding to the given sample
    atlas_info : pandas.DataFrame,
        Dataframe containing information about the atlas of interest. Must have
        _at least_ columns 'id', 'hemisphere', and 'structure' containing
        information mapping atlas IDs to hemisphere and broad structural class
        (i.e., "cortex", "subcortex", "cerebellum"). Default: None

    Returns
    -------
    label : int
        New label for sample
    """

    cols = ['hemisphere', 'structure']

    if label != 0:
        sample_info = sample_info[cols]
        atlas_info = atlas_info.loc[label, cols]
        if not np.all(np.asarray(sample_info) == np.asarray(atlas_info)):
            label = 0

    return label


def label_samples(annotation, atlas, atlas_info=None, tolerance=2):
    """
    Matches all microarray samples in `annotation` to parcels in `atlas`

    Attempts to place each sample provided in `annotation` into a parcel in
    `atlas`, where the latter is a 3D niimg-like object that contains parcels
    each idnetified by a unique integer ID.

    The function tries to best match samples in `annotation` to parcels defined
    in `atlas` by:

        1. Determining if the sample falls directly within a parcel,
        2. Checking to see if there are nearby parcels by slowly expanding the
           search space to include nearby voxels, up to a specified distance
           (specified via the `tolerance` parameter),
        3. Assigning the sample to the closest parcel if there are multiple
           nearby parcels, where closest is determined by the parcel centroid.

    If at any step a sample can be assigned to a parcel the matching process is
    terminated. If there is still no parcel for a given sample after this
    process the sample is provided a label of 0.

    Parameters
    ----------
    annotation : (S, 13) pandas.DataFrame
        Pre-loaded annotation information for a given AHBA donor
    atlas : niimg-like object
        A parcellation image in MNI space, where each parcel is identified by a
        unique integer ID
    atlas_info : pandas.DataFrame, optional
        Filepath to or pre-loaded dataframe containing information about
        `atlas`. Must have *at least* columns 'id', 'hemisphere', and
        'structure' containing information mapping atlas IDs to hemisphere and
        broad structural class (i.e., "cortex", "subcortex", "cerebellum").
        Default: None
    tolerance : int, optional
        Distance (in mm) that a sample must be from a parcel for it to be
        matched to that parcel. This is only considered if the sample is not
        directly within a parcel. Default: 2

    Returns
    -------
    labels : (S, 1) pandas.DataFrame
        Dataframe with parcel labels for each of `S` samples
    """

    # get annotation and atlas data
    annotation = io.read_annotation(annotation)
    atlas = utils.check_img(atlas)
    label_data, affine = np.asarray(atlas.dataobj), atlas.affine

    # load atlas_info, if provided
    if atlas_info is not None:
        atlas_info = utils.check_atlas_info(atlas, atlas_info)

    # get ijk coordinates for microarray samples and find labels
    g_ijk = utils.xyz_to_ijk(annotation[['mni_x', 'mni_y', 'mni_z']], affine)
    labelled = label_data[g_ijk[:, 0], g_ijk[:, 1], g_ijk[:, 2]]

    # if sample coordinates aren't directly inside a parcel, increment radius
    # around sample up to `tolerance` to try and find nearby parcels.
    # if still no parcel, then ignore this sample
    for idx in np.where(labelled == 0)[0]:
        label, tol = labelled[idx], 1
        while label == 0 and tol <= tolerance:
            label = _assign_sample(g_ijk[[idx]], atlas,
                                   sample_info=annotation.iloc[idx],
                                   atlas_info=atlas_info,
                                   tolerance=tol)
            tol += 1
        labelled[idx] = label

    return pd.DataFrame(labelled, dtype=int,
                        columns=['label'], index=annotation.index)


def groupby_index(microarray, labels=None, metric='mean'):
    """
    Averages expression data in `microarray` over samples with same label

    Parameters
    ----------
    microarray : (S, G) pandas.DataFrame
        Microarray expression data, where `S` is samples and `G` is genes
    labels : (L,) array_like, optional
        All possible labels for parcellation (to account for possibility that
        some parcels have NO expression data). Default: None
    metric : str or func, optional
        Mechanism by which to collapse across samples within a parcel. If a
        str, should be in ['mean', 'median']; if a function, should be able to
        accept an `N`-dimensional input and the `axis` keyword argument and
        return an `N-1`-dimensional output. Default: 'mean'

    Returns
    -------
    gene_by_label : (L, G) pandas.DataFrame
        Microarray expression data
    """

    # get combination function
    metric = utils.check_metric(metric)

    # get missing labels
    if labels is not None:
        missing = np.setdiff1d(labels, np.unique(microarray.index))
        labels = pd.DataFrame(columns=microarray.columns,
                              index=pd.Series(missing, name='label'))

    gene_by_label = (microarray.groupby('label')
                               .aggregate(metric)
                               .append(labels)
                               .sort_index()
                               .rename_axis('label'))

    # remove "zero" label (if it exists)
    if 0 in gene_by_label.index:
        gene_by_label = gene_by_label.drop([0], axis=0)

    return gene_by_label


def aggregate_samples(microarray, labels=None, region_agg='donors',
                      agg_metric='mean', return_donors=False):
    """
    Aggregates samples in `microarray` belonging to same regions

    Parameters
    ----------
    microarray : list of (S, G) pandas.DataFrame
        Microarray expression data for multiple donor, where `S` is samples and
        `G` is genes. Index of dataframes should identify to which region each
        sample was assigned
    labels : array_like
        Array containing IDs of all possible region labels. If no samples from
        any donor were assigned to a region NaNs are inserted in place of
        expression values
    region_agg : {'samples', 'donors'}, optional
        When multiple samples are identified as belonging to a region in
        `atlas` this determines how they are aggegated. If 'samples',
        expression data from all samples for all donors assigned to a given
        region are combined. If 'donors', expression values for all samples
        assigned to a given region are combined independently for each donor
        before being combined across donors. See `agg_metric` for mechanism by
        which samples are combined. Default: 'donors'
    agg_metric : {'mean', 'median'} or callable, optional
        Mechanism by which to reduce sample-level expression data into region-
        level expression (see `region_agg`). If a callable, should be able to
        accept an `N`-dimensional input and the `axis` keyword argument and
        return an `N-1`-dimensional output. Default: 'mean'
    return_donors : bool, optional
        Whether to return donor-level expression arrays instead of aggregating
        expression across donors with provided `agg_metric`. Default: False

    Returns
    -------
    microarray : (R, G) pandas.DataFrame or list of (R, G) pandas.DataFrame
        Microarray expression data aggregated across samples within each of `R`
        regions. If `return_donors=True` then a list of dataframes is returned
    """

    region_aggs = ['donors', 'samples']
    if region_agg not in region_aggs:
        raise ValueError('Provided `region_agg` {} invalid. Must be one of {}'
                         .format(region_agg, region_aggs))
    if region_agg == 'samples' and return_donors:
        raise ValueError('Cannot use region_agg=\'samples\' when '
                         '`return_donors=True`.')
    metric = utils.check_metric(agg_metric)

    lgr.info('Aggregating samples to regions with provided region_agg: {}'
             .format(region_agg))

    if region_agg == 'donors':
        microarray = [
            groupby_index(e, labels=labels, metric=metric) for e in microarray
        ]
    elif region_agg == 'samples':
        microarray = [
            groupby_index(pd.concat(microarray), labels=labels, metric=metric)
        ]

    if not return_donors:
        microarray = pd.concat(microarray).groupby('label').aggregate(metric)
        # some genes may have been poorly normalized; remove these
        drop = microarray.dropna(axis=0, how='all').isna().any(axis=0)
        lgr.info('Dropping {} gene from concatenated expression data due to '
                 'poor normalization'.format(drop.sum()))
        microarray = microarray.drop(drop[drop].index, axis=1)

    return microarray



# -*- coding: utf-8 -*-
"""
Functions for mapping AHBA microarray dataset to atlases and and parcellations
"""

from functools import reduce

import nibabel as nib
import numpy as np
import pandas as pd

from . import correct, datasets, io, probes_, samples_, utils
from .utils import first_entry, flatten_dict

import logging
logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
lgr = logging.getLogger('abagen')
lgr_levels = dict(zip(range(3), [40, 20, 10]))


def get_expression_data(atlas,
                        atlas_info=None,
                        *,
                        ibf_threshold=0.5,
                        probe_selection='diff_stability',
                        donor_probes='aggregate',
                        lr_mirror=False,
                        exact=True,
                        tolerance=2,
                        sample_norm='srs',
                        gene_norm='srs',
                        norm_matched=True,
                        region_agg='donors',
                        agg_metric='mean',
                        corrected_mni=True,
                        reannotated=True,
                        return_counts=False,
                        return_donors=False,
                        donors='all',
                        data_dir=None,
                        verbose=1,
                        n_proc=1):
    """
    Assigns microarray expression data to ROIs defined in `atlas`

    This function aims to provide a workflow for generating pre-processed,
    microarray expression data from the Allen Human Brain Atlas ([A2]_) for
    abitrary `atlas` designations. First, some basic filtering of genetic
    probes is performed, including:

        1. Intensity-based filtering of microarray probes to remove probes that
           do not exceed a certain level of background noise (specified via the
           `ibf_threshold` parameter), and
        2. Selection of a single, representative probe (or collapsing across
           probes) for each gene, specified via the `probe_selection`
           parameter.

    Tissue samples are then matched to parcels in the defined `atlas` for each
    donor. If `atlas_info` is provided then this matching is constrained by
    both hemisphere and tissue class designation (e.g., cortical samples from
    the left hemisphere are only matched to ROIs in the left cortex,
    subcortical samples from the right hemisphere are only matched to ROIs in
    the left subcortex); see the `atlas_info` parameter description for more
    information.

    Matching of microarray samples to parcels in `atlas` is done via a multi-
    step process:

        1. Determine if the sample falls directly within a parcel,
        2. Check to see if there are nearby parcels by slowly expanding the
           search space to include nearby voxels, up to a specified distance
           (specified via the `tolerance` parameter),
        3. If there are multiple nearby parcels, the sample is assigned to the
           closest parcel, as determined by the parcel centroid.

    If at any step a sample can be assigned to a parcel the matching process is
    terminated. If multiple sample are assigned to the same parcel they are
    aggregated with the metric specified via the `metric` parameter. More
    control over the sample matching can be obtained by setting the `exact`
    parameter; see the parameter description for more information.

    Once all samples have been matched to parcels for all supplied donors, the
    microarray expression data are optionally normalized via the provided
    `sample_norm` and `gene_norm` functions before being combined within
    parcels and across donors via the supplied `agg_metric`.

    Parameters
    ----------
    atlas : niimg-like object or dict
        A parcellation image in MNI space, where each parcel is identified by a
        unique integer ID. Alternatively, a dictionary where keys are donor IDs
        and values are parcellation images in the native space of each donor.
    atlas_info : os.PathLike or pandas.DataFrame, optional
        Filepath to or pre-loaded dataframe containing information about
        `atlas`. Must have at least columns 'id', 'hemisphere', and 'structure'
        containing information mapping atlas IDs to hemisphere (i.e, "L", "R")
        and broad structural class (i.e., "cortex", "subcortex", "cerebellum").
        If provided, this will constrain matching of tissue samples to regions
        in `atlas`. Default: None
    ibf_threshold : [0, 1] float, optional
        Threshold for intensity-based filtering. This number specifies the
        ratio of samples, across all supplied donors, for which a probe must
        have signal significantly greater background noise in order to be
        retained. Default: 0.5
    probe_selection : str, optional
        Selection method for subsetting (or collapsing across) probes that
        index the same gene. Must be one of 'average', 'max_intensity',
        'max_variance', 'pc_loading', 'corr_variance', 'corr_intensity', or
        'diff_stability', 'rnaseq'; see Notes for more information on different
        options. Default: 'diff_stability'
    donor_probes : str, optional
        Whether specified `probe_selection` method should be performed with
        microarray data from all donors ('aggregate'), independently for each
        donor ('independent'), or based on the most common selected probe
        across donors ('common'). Not all combinations of `probe_selection`
        and `donor_probes` methods are viable. Default: 'aggregate'
    lr_mirror : bool, optional
        Whether to mirror microarray expression samples across hemispheres to
        increase spatial coverage. This will duplicate samples across both
        hemispheres (i.e., L->R and R->L), approximately doubling the number of
        available samples. Default: False
    exact : bool, optional
        Whether to use exact matching of donor tissue samples to parcels in
        `atlas`. If True, this function will ONLY match tissue samples to
        parcels within `threshold` mm of the sample; any samples that are
        beyond `threshold` mm of a parcel will be discarded. This may result
        in some parcels having no assigned sample / expression data. If False,
        the default matching procedure will be performed and followed by a
        check for parcels with no assigned samples; any such parcels will be
        matched to the nearest sample (defined as the sample with the closest
        Euclidean distance to the parcel centroid). Default: True
    tolerance : int, optional
        Distance (in mm) that a sample must be from a parcel for it to be
        matched to that parcel. This is only considered if the sample is not
        directly within a parcel. Default: 2
    sample_norm : {'rs', 'srs', 'minmax', 'center', 'zscore', None}, optional
        Method by which to normalize microarray expression values for each
        sample. Expression values are normalized separately for each sample and
        donor across all genes; see Notes for more information on different
        methods. If None is specified then no normalization is performed.
        Default: 'srs'
    gene_norm : {'rs', 'srs', 'minmax', 'center', 'zscore', None}, optional
        Method by which to normalize microarray expression values for each
        donor. Expression values are normalized separately for each gene and
        donor across all samples; see Notes for more information on different
        methods. If None is specified then no normalization is performed.
        Default: 'srs'
    norm_matched : bool, optional
        Whether to perform gene normalization (`gene_norm`) across only those
        samples matched to regions in `atlas` instead of all available samples.
        If `atlas` is very small (i.e., only a few regions of interest), using
        `norm_matched=False` is suggested. Default: True
    region_agg : {'samples', 'donors'}, optional
        When multiple samples are identified as belonging to a region in
        `atlas` this determines how they are aggegated. If 'samples',
        expression data from all samples for all donors assigned to a given
        region are combined. If 'donors', expression values for all samples
        assigned to a given region are combined independently for each donor
        before being combined across donors. See `agg_metric` for mechanism by
        which samples are combined. Default: 'donors'
    agg_metric : {'mean', 'median'} or callable, optional
        Mechanism by which to reduce sample-level expression data into region-
        level expression (see `region_agg`). If a callable, should be able to
        accept an `N`-dimensional input and the `axis` keyword argument and
        return an `N-1`-dimensional output. Default: 'mean'
    corrected_mni : bool, optional
        Whether to use the "corrected" MNI coordinates shipped with the
        `alleninf` package instead of the coordinates provided with the AHBA
        data when matching tissue samples to anatomical regions. Default: True
    reannotated : bool, optional
        Whether to use reannotated probe information provided by [A1]_ instead
        of the default probe information from the AHBA dataset. Using
        reannotated information will discard probes that could not be reliably
        matched to genes. Default: True
    return_counts : bool, optional
        Whether to return dataframe containing information on how many samples
        were assigned to each parcel in `atlas` for each donor. Default: False
    return_donors : bool, optional
        Whether to return donor-level expression arrays instead of aggregating
        expression across donors with provided `agg_metric`. Default: False
    donors : list, optional
        List of donors to use as sources of expression data. Can be either
        donor numbers or UID. If not specified will use all available donors.
        Note that donors '9861' and '10021' have samples from both left + right
        hemispheres; all other donors have samples from the left hemisphere
        only. Default: 'all'
    data_dir : os.PathLike, optional
        Directory where expression data should be downloaded (if it does not
        already exist) / loaded. If not specified will use the current
        directory. Default: None
    verbose : int, optional
        Specifies verbosity of status messages to display during workflow.
        Higher numbers increase verbosity of messages while zero suppresses all
        messages. Default: 1
    n_proc : int, optional
        Number of processors to use to download AHBA data. Can parallelize up
        to six times. Default: 1

    Returns
    -------
    expression : (R, G) pandas.DataFrame
        Microarray expression for `R` regions in `atlas` for `G` genes,
        aggregated across donors, where the index corresponds to the unique
        integer IDs of `atlas` and the columns are gene names. If
        ``return_donors=True`` then this is a list of (R, G) dataframes, one
        for each donor.
    counts : (R, D) pandas.DataFrame
        Number of samples assigned to each of `R` regions in `atlas` for each
        of `D` donors (if multiple donors were specified); only returned if
        ``return_counts=True``.

    Notes
    -----
    The following methods can be used for collapsing across probes when
    multiple probes are available for the same gene:

        1. ``probe_selection='average'``

        Takes the average of expression data across all probes indexing the
        same gene. Providing 'mean' as the input method will return the same
        thing. This method can only be used when `donor_probes='aggregate'`.

        2. ``probe_selection='max_intensity'``

        Selects the probe with the maximum average expression across samples
        from all donors.

        3. ``probe_selection='max_variance'``

        Selects the probe with the maximum variance in expression across
        samples from all donors.

        4. ``probe_selection='pc_loading'``

        Selects the probe with the maximum loading along the first principal
        component of a decomposition performed across samples from all donors.

        5. ``probe_selection='corr_intensity'``

        Selects the probe with the maximum correlation to other probes from the
        same gene when >2 probes exist; otherwise, uses the same procedure as
        `max_intensity`.

        6. ``probe_selection='corr_variance'``

        Selects the probe with the maximum correlation to other probes from the
        same gene when >2 probes exist; otherwise, uses the same procedure as
        `max_varance`.

        7. ``probe_selection='diff_stability'``

        Selects the probe with the most consistent pattern of regional
        variation across donors (i.e., the highest average correlation across
        brain regions between all pairs of donors). This method can only be
        used when `donor_probes='aggregate'`.

        8. ``method='rnaseq'``

        Selects probes with most consistent pattern of regional variation to
        RNAseq data (across the two donors with RNAseq data). This method can
        only be used when `donor_probes='aggregate'`.

    Note that for incompatible combinations of `probe_selection` and
    `donor_probes` (as detailed above), the `probe_selection choice will take
    precedence. For example, providing ``probe_selection='diff_stability'`` and
    ``donor_probes='independent'`` will cause `donor_probes` to be reset to
    `'aggregate'`.

    The following methods can be used for normalizing microarray expression
    values prior to aggregating:

        1. ``{sample,gene}_norm=='rs'``

        Uses a robust sigmoid function as in [A3]_ to normalize values

        2. ``{sample,gene}_norm='srs'``

        Same as 'rs' but scales output to the unit normal (i.e., range 0-1)

        3. ``{sample,gene}_norm='minmax'``

        Scales data to the unit normal (i.e., range 0-1)

        4. ``{sample,gene}_norm='center'``

        Removes the mean of expression values

        5. ``{sample,gene}_norm='zscore'``

        Applies a basic z-score (subtract mean, divide by standard deviation);
        uses degrees of freedom equal to one for standard deviation

    References
    ----------
    .. [A1] Arnatkevic̆iūtė, A., Fulcher, B. D., & Fornito, A. (2019). A
       practical guide to linking brain-wide gene expression and neuroimaging
       data. NeuroImage, 189, 353-367.
    .. [A2] Hawrylycz, M.J. et al. (2012) An anatomically comprehensive atlas
       of the adult human transcriptome. Nature, 489, 391-399.
    .. [A3] Fulcher, B. D., & Fornito, A. (2016). A transcriptional signature
       of hub connectivity in the mouse connectome. Proceedings of the National
       Academy of Sciences, 113(5), 1435-1440.
    """

    # set logging verbosity level
    lgr.setLevel(lgr_levels.get(int(verbose), 1))

    # load atlas and atlas_info, if provided, and coerce to dict
    atlas, atlas_info, same = coerce_atlas_to_dict(atlas, donors, atlas_info)

    # get combination functions
    agg_metric = utils.check_metric(agg_metric)

    # check probe_selection input
    if probe_selection not in probes_.SELECTION_METHODS:
        raise ValueError('Provided probe_selection method is invalid, must be '
                         f'one of {list(probes_.SELECTION_METHODS)}. Received '
                         f'value: \'{probe_selection}\'')
    if donor_probes not in ['aggregate', 'independent', 'common']:
        raise ValueError('Provided donor_probes method is invalid, must be '
                         f'one of [\'aggregate\', \'independent\']. Received '
                         f'value: \'{donor_probes}\'')

    # fetch files (downloading if necessary) and unpack to variables
    files = datasets.fetch_microarray(data_dir=data_dir, donors=donors,
                                      verbose=verbose, n_proc=n_proc)

    if probe_selection == 'diff_stability' and len(files) == 1:
        raise ValueError('Cannot use diff_stability for probe_selection with '
                         'only one donor. Please specify a different probe_'
                         'selection method or use more donors.')
    elif probe_selection == 'rnaseq':  # fetch RNAseq if we're gonna need it
        datasets.fetch_rnaseq(data_dir=data_dir, donors=donors,
                              verbose=verbose)

    # get some info on labels in `atlas_img`
    all_labels = utils.get_unique_labels(utils.first_entry(atlas))
    n_gb = (8 * len(all_labels) * 30000) / (1024 ** 3)
    if n_gb > 1:
        lgr.warning(f'Output region x gene matrix may require up to {n_gb:.2f}'
                    'GB RAM.')
    if not exact and same:
        lgr.info(f'Pre-calculating centroids for {len(all_labels)} regions in '
                 'provided atlas')
        centroids = utils.get_centroids(utils.first_entry(atlas),
                                        labels=all_labels,
                                        image_space=True)

    # update the annotation "files". this handles updating the MNI coordinates,
    # dropping mistmatched samples (where MNI coordinates don't match the
    # provided ontology), and mirroring samples across hemispheres, if desired
    for donor, data in files.items():
        annot, ontol = data['annotation'], data['ontology']
        t1w = None
        if not same:
            t1w = datasets.fetch_raw_mri(donors=donor,
                                         data_dir=data_dir,
                                         verbose=verbose)[donor]['t1w']
        annot = samples_.update_coords(annot, corrected_mni=corrected_mni,
                                       native_space=t1w)
        annot = samples_.drop_mismatch_samples(annot, ontol)
        if lr_mirror:
            annot = samples_.mirror_samples(annot, ontol)
        data['annotation'] = annot
    annotation = flatten_dict(files, 'annotation')

    # get dataframe of probe information (reannotated or otherwise)
    # the Probes.csv files are the same for every donor so just grab the first
    probe_info = io.read_probes(first_entry(files, 'probes'))
    if reannotated:
        probe_info = probes_.reannotate_probes(probe_info)

    # drop probes with no/invalid Entrez ID
    probe_info = probe_info.dropna(subset=['entrez_id'])

    # intensity-based filtering of probes
    probe_info = probes_.filter_probes(flatten_dict(files, 'pacall'),
                                       annotation, probe_info,
                                       threshold=ibf_threshold)

    # get probe-reduced microarray expression data for all donors based on
    # selection method; this will be a list of gene x sample dataframes (one
    # for each donor)
    microarray = probes_.collapse_probes(flatten_dict(files, 'microarray'),
                                         annotation, probe_info,
                                         method=probe_selection,
                                         donor_probes=donor_probes)
    missing = []
    counts = pd.DataFrame(np.zeros((len(all_labels) + 1, len(microarray)),
                                   dtype=int),
                          index=np.append([0], all_labels),
                          columns=microarray.keys())
    for subj in microarray:
        if lr_mirror:  # reset index (duplicates will cause issues if we don't)
            # TODO: come up with alternative sample IDs for mirrored samples
            annotation[subj] = annotation[subj].reset_index(drop=True)
            microarray[subj] = microarray[subj].reset_index(drop=True)

        # assign samples to regions
        labels = samples_.label_samples(annotation[subj], atlas[subj],
                                        atlas_info, tolerance=tolerance)

        # if we're doing exact matching and want to aggregate samples w/i
        # regions, remove the non-labelled samples prior to normalization.
        # otherwise, we'll remove the non-labelled samples after normalization
        nz = np.asarray(labels != 0).squeeze()
        if norm_matched:
            microarray[subj] = microarray[subj].loc[nz]
            annotation[subj] = annotation[subj].loc[nz]
            labels = labels.loc[nz]

        if sample_norm is not None:
            microarray[subj] = correct.normalize_expression(microarray[subj].T,
                                                            norm=sample_norm,
                                                            ignore_warn=True).T
        if gene_norm is not None:
            microarray[subj] = correct.normalize_expression(microarray[subj],
                                                            norm=gene_norm,
                                                            ignore_warn=True)

        # get counts of samples collapsed into each ROI
        labs, num = np.unique(labels, return_counts=True)
        counts.loc[labs, subj] = num
        lgr.info(f'{counts.iloc[1:][subj].sum():>3} / {len(nz)} '
                 f'samples matched to regions for donor #{subj}')

        # if we don't want to do exact matching then cache which parcels are
        # missing data and the expression data for the closest sample to that
        # parcel; we'll use this once we've iterated through all donors
        if not exact:
            if not same:
                centroids = utils.get_centroids(atlas[subj], labels=all_labels,
                                                image_space=True)
            empty = np.logical_not(np.in1d(all_labels, labs))
            cols = ['mni_x', 'mni_y', 'mni_z']
            idx, dist = utils.closest_centroid(annotation[subj][cols],
                                               centroids[empty],
                                               return_dist=True)
            if not hasattr(idx, '__len__'):  # TODO: better way to check this?
                idx, dist = np.array([idx]), np.array([dist])
            idx = microarray[subj].loc[annotation[subj].iloc[idx].index]
            empty = all_labels[empty]
            idx.index = pd.Series(empty, name='label')
            missing += [(idx, dict(zip(empty, np.diag(dist))))]

        microarray[subj].index = labels['label']

    if not exact:  # check for missing ROIs and fill in, as needed
        # labels that are missing across all donors
        empty = reduce(set.intersection, [set(f.index) for f, d in missing])
        lgr.info(f'Matching {len(empty)} regions w/no data to nearest samples')
        for roi in empty:
            # find donor with sample closest to centroid of empty parcel
            ind = np.argmin([dist.get(roi) for micro, dist in missing])
            subj = list(microarray.keys())[ind]
            lgr.debug(f'Assigning sample from donor {subj} to region #{roi}')
            # assign expression data from that sample and add to count
            exp = missing[ind][0].loc[roi]
            microarray[subj] = microarray[subj].append(exp)
            counts.loc[roi, subj] += 1

    # if we don't want to aggregate over regions return voxel-level results
    if region_agg is None:
        # don't return samples that aren't matched to a region in the `atlas`
        mask = {d: m.index != 0 for d, m in microarray.items()}
        microarray = pd.concat([m[mask[d]] for d, m in microarray.items()])
        # set index to well_id for all remaining tissue samples
        microarray.index = pd.Series(np.asarray(
            pd.concat([a[mask[d]] for d, a in annotation.items()])['well_id']
        ), name='well_id')
        # return expression data (remove NaNs)
        return microarray.dropna(axis=1, how='any')

    microarray = samples_.aggregate_samples(microarray.values(),
                                            labels=all_labels,
                                            region_agg=region_agg,
                                            agg_metric=agg_metric,
                                            return_donors=return_donors)

    # drop the "zero" label from the counts dataframe (this is background)
    if return_counts:
        return microarray, counts.iloc[1:]

    return microarray


def get_samples_in_mask(mask=None, **kwargs):
    """
    Returns preprocessed microarray expression data for samples in `mask`

    Uses the same processing workflow as :func:`abagen.get_expression_data` but
    instead of aggregating samples within regions simply returns sample-level
    expression data for all samples that fall within boundaries of `mask`.

    Parameters
    ----------
    mask : niimg-like object, optional
        A mask image in MNI space (where 0 is the background). Alternatively, a
        dictionary where keys are donor IDs and values are mask images in the
        native space of each donor. If not supplied, all available samples will
        be returned. Default: None
    kwargs : key-value pairs
        All key-value pairs from :func:`abagen.get_expression_data` except for:
        `atlas`, `atlas_info`, `region_agg`, and `agg_metric`, which will be
        ignored. If `atlas` is supplied instead of `mask` then `atlas` will be
        used instead as a modified binary image. If both `atlas` and `mask` are
        supplied then `mask` will be usedin

    Returns
    -------
    expression : (S, G) pandas.DataFrame
        Microarray expression for `S` samples for `G` genes, aggregated across
        donors, where the columns are gene names
    coords : (S,) numpy.ndarray
        MNI coordinates of samples in `expression`. Even if donor-specific
        masks are provided MNI coordinates will be returned to ensure
        comparability between subjects
    """

    # fetch files (downloading if necessary) to get coordinates
    files = datasets.fetch_microarray(data_dir=kwargs.get('data_dir', None),
                                      donors=kwargs.get('donors', 'all'),
                                      verbose=kwargs.get('verbose', 1),
                                      n_proc=kwargs.get('n_proc', 1))

    # get updated coordinates
    for donor, data in files.items():
        annot, ontol = data['annotation'], data['ontology']
        if kwargs.get('corrected_mni', True):
            annot = samples_.update_mni_coords(annot)
        annot = samples_.drop_mismatch_samples(annot, ontol)
        if kwargs.get('lr_mirror', False):
            annot = samples_.mirror_samples(annot, ontol)
        data['annotation'] = annot
    cols = ['well_id', 'mni_x', 'mni_y', 'mni_z']
    coords = np.asarray(pd.concat(flatten_dict(files, 'annotation'))[cols])
    well_id, coords = np.asarray(coords[:, 0], 'int'), coords[:, 1:]

    # in case people mix things up and use atlas instead of mask, use that
    if kwargs.get('atlas') is not None and mask is None:
        mask = kwargs['atlas']
    elif mask is None:
        # create affine for "full" mask
        affine = np.eye(4)
        affine[:-1, -1] = np.floor(coords).min(axis=0)

        # downsample coordinates to specified resolution and convert to ijk
        ijk = np.unique(
            np.asarray(
                np.floor(
                    nib.affines.apply_affine(np.linalg.inv(affine), coords),
                ), dtype='int'
            ), axis=0
        )

        # generate atlas image where each voxel has
        img = np.zeros(ijk.max(axis=0) + 2, dtype='int')
        img[tuple(map(tuple, ijk.T))] = 1
        mask = nib.Nifti1Image(img, affine=affine)

    # reset these parameters
    kwargs['atlas'] = mask
    kwargs['atlas_info'] = None
    kwargs['region_agg'] = None
    # soft reset this parameter
    kwargs.setdefault('norm_matched', False)

    # get expression data + drop sample coordinates that weren't in atlas
    exp = get_expression_data(**kwargs)
    coords = coords[np.isin(well_id, exp.index)]

    return exp, coords


def coerce_atlas_to_dict(atlas, donors, atlas_info=None):
    """
    Coerces `atlas` to dict with keys `donors`

    If already a dictionary, confirms that `atlas` has entries for all values
    in `donors`

    Parameters
    ----------
    atlas : niimg-like object
        A parcellation image in MNI space, where each parcel is identified by a
        unique integer ID
    donors : array_like
        Donors that should have entries in returned `atlas` dictionary
    atlas_info : os.PathLike or pandas.DataFrame, optional
        Filepath to or pre-loaded dataframe containing information about
        `atlas`. Must have at least columns 'id', 'hemisphere', and 'structure'
        containing information mapping atlas IDs to hemisphere (i.e, "L", "R")
        and broad structural class (i.e., "cortex", "subcortex", "cerebellum").
        If provided, this will constrain matching of tissue samples to regions
        in `atlas`. Default: None

    Returns
    -------
    atlas : dict
        Dict where keys are `donors` and values are `atlas`. If a dict was
        provided it is checked to ensure
    atlas_info : pandas.DataFrame
        Loaded dataframe with information on atlas
    same : bool
        Whether one atlas was provided for all donors (True) instead of
        donor-specific atlases (False)
    """

    donors = datasets.fetchers.check_donors(donors)
    same = True

    # FIXME: so that we're not depending on type checks so much :grimacing:
    if isinstance(atlas, dict):
        atlas = {
            datasets.WELL_KNOWN_IDS.subj[d]: utils.check_img(a)
            for d, a in atlas.items()
        }
        same = False
        missing = set(donors) - set(atlas)
        if len(missing) > 0:
            raise ValueError('Provided `atlas` does not have entry for all '
                             f'requested donors. Missing donors: {donors}.')
        lgr.info('Donor-specific atlases provided; using native MRI '
                 'coordinates for tissue samples')
    else:
        atlas = utils.check_img(atlas)
        atlas = {donor: atlas for donor in donors}
        lgr.info('Group-level atlas provided; using MNI coordinates for '
                 'tissue samples')

    if atlas_info is not None:
        for donor, atl in atlas.items():
            atlas_info = utils.check_atlas_info(atlas[donor], atlas_info)

    return atlas, atlas_info, same



__all__ = [
    '__version__', '__doc__',
    'io', 'mouse', 'get_expression_data', 'get_samples_in_mask'
    'keep_stable_genes', 'normalize_expression', 'remove_distance',
    'fetch_desikan_killiany', 'fetch_gene_group', 'fetch_microarray',
    'fetch_raw_mri', 'fetch_rnaseq', 'fetch_freesurfer', 'fetch_donor_info'
]

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .info import long_description as __doc__

from . import io, mouse
from .allen import get_expression_data, get_samples_in_mask
from .correct import keep_stable_genes, normalize_expression, remove_distance
from .datasets import (fetch_desikan_killiany, fetch_gene_group,
                       fetch_microarray, fetch_raw_mri, fetch_rnaseq,
                       fetch_freesurfer, fetch_donor_info)



# -*- coding: utf-8 -*-
"""
Tests for abagen.correct module
"""

import itertools

import numpy as np
import pandas as pd
import pytest
import scipy.stats as sstats

from abagen import allen, correct, io
from abagen.utils import flatten_dict


@pytest.fixture(scope='module')
def donor_expression(testfiles, atlas):
    return allen.get_expression_data(atlas['image'], atlas['info'],
                                     exact=False, return_donors=True,
                                     donors=['12876', '15496'])


def test__unpack_tuple():
    assert correct._unpack_tuple((3,)) == 3
    assert correct._unpack_tuple((3, 3)) == (3, 3)
    assert correct._unpack_tuple([2]) == 2
    assert correct._unpack_tuple([2, 4]) == [2, 4]
    assert correct._unpack_tuple(np.array([3])) == 3
    assert np.all(correct._unpack_tuple(np.array([3, 3])) == [3, 3])


def test__batch():
    rs = np.random.RandomState(1234)
    # p-values for ANOVA should all be ~0 (large group differences) before
    # batch correction
    y = [rs.normal(size=(100, 1000)) + f for f in [5, 0, 0]]
    assert np.allclose(sstats.f_oneway(*y)[1], 0)

    # F-values for ANOVA should all be ~0 (no group differences) after batch
    # correction; p-values returned here are sometimes NaN so not a good test
    out = correct._batch_correct(y)
    assert np.allclose(sstats.f_oneway(*out)[0], 0)

    # mean expressions after correction should be ~equal
    assert np.allclose([o.mean() for o in out], 1.24871965683026)

    with pytest.raises(ValueError):
        correct._batch_correct([y[0]])


def test__rescale():
    rs = np.random.RandomState(1234)
    y = rs.normal(size=(100, 1000)) + 10
    out = correct._rescale(y)

    # default max = 1, min =0
    assert np.allclose(out.max(axis=0), 1) and np.allclose(out.min(axis=0), 0)

    # can specify alternative min/max
    out = correct._rescale(y, low=5, high=6)
    assert np.allclose(out.max(axis=0), 6) and np.allclose(out.min(axis=0), 5)

    # different axis works, too!
    out = correct._rescale(y, axis=1)
    assert np.allclose(out.max(axis=1), 1) and np.allclose(out.min(axis=1), 0)


@pytest.mark.parametrize('a', [0, 1])
def test__rs(a):
    rs = np.random.RandomState(1234)

    # create an array with a pretty ridiculous outlier effect to try and fix
    y = rs.normal(size=(100, 1000))
    y[0] += 1000
    y[:, 0] += 1000
    out = correct._rs(y, axis=a)

    # max will always be less than one, min will always be greater than zero
    assert np.all(out.max(axis=a) <= 1) and np.all(out.min(axis=a) >= 0)

    # we should have reduced skewness / kurtosis compared to the original
    assert np.all(sstats.skew(out, axis=a) < sstats.skew(y, axis=a))
    assert np.all(sstats.kurtosis(out, axis=a) < sstats.kurtosis(y, axis=a))

    # this is a weird test; we're gonna bin the data at 0.2 intervals and make
    # sure no bins are empty. if one is something probably went wrong, right?
    for low in np.arange(0, 1, 0.2):
        hi = low + 0.2 + np.spacing(1)  # include 1
        assert np.all(np.sum(np.logical_and(out >= low, out < hi), axis=a) > 0)


@pytest.mark.parametrize('a', [0, 1])
def test__srs(a):
    rs = np.random.RandomState(1234)

    # create an array with a pretty ridiculous outlier effect to try and fix
    y = rs.normal(size=(100, 1000))
    y[0] += 1000
    y[:, 0] += 1000
    out = correct._srs(y, axis=a)

    # max will always be one, min will always be zero
    assert np.allclose(out.max(axis=a), 1) and np.allclose(out.min(axis=a), 0)

    # we should have reduced skewness / kurtosis compared to the original
    assert np.all(sstats.skew(out, axis=a) < sstats.skew(y, axis=a))
    assert np.all(sstats.kurtosis(out, axis=a) < sstats.kurtosis(y, axis=a))

    # this is a weird test; we're gonna bin the data at 0.2 intervals and make
    # sure no bins are empty. if one is something probably went wrong, right?
    for low in np.arange(0, 1, 0.2):
        hi = low + 0.2 + np.spacing(1)  # include 1
        assert np.all(np.sum(np.logical_and(out >= low, out < hi), axis=a) > 0)


@pytest.mark.parametrize('method', [
    'center', 'zscore', 'minmax', 'sigmoid', 'scaled_sigmoid',
    'scaled_sigmoid_quantiles', 'robust_sigmoid', 'scaled_robust_sigmoid',
    'mixed_sigmoid'
])
def test_normalize_expression_real(testfiles, method):
    # load in data and add some NaN values for "realness"
    micro = [
        io.read_microarray(f).T
        for f in flatten_dict(testfiles, 'microarray').values()
    ]
    inds = [[5, 15, 25], [0, 10, 20]]
    for n, idx in enumerate(inds):
        micro[n].iloc[idx] = np.nan

    minmax = [
        'minmax', 'scaled_sigmoid', 'scaled_sigmoid_quantiles',
        'scaled_robust_sigmoid', 'mixed_sigmoid'
    ]

    out = correct.normalize_expression(micro, norm=method)
    for exp, idx in zip(out, inds):
        assert np.all(np.isnan(exp.iloc[idx]))
        exp = exp.dropna(axis=1, how='all')
        if method in minmax:
            assert np.allclose(exp.max(axis=0), 1)
            assert np.allclose(exp.min(axis=0), 0)
        elif method == 'robust_sigmoid':
            assert np.all(exp.max(axis=0) <= 1)
            assert np.all(exp.min(axis=0) >= 0)
        elif method in ['center', 'zscore']:
            assert np.allclose(exp.mean(axis=0), 0)
            if method == 'zscore':
                assert np.allclose(exp.std(axis=0, ddof=1), 1)

    # # batch correct: force means identical
    # out = correct.normalize_expression(micro, norm='batch')
    # assert np.allclose(*[e.mean(axis=0, skipna=True) for e in out])
    # # the NaN values should still be there, though
    # for exp, idx in zip(out, inds):
    #     assert np.all(np.isnan(exp.iloc[idx]))

    # invalid norm parameter
    with pytest.raises(ValueError):
        correct.normalize_expression(micro, norm='notanorm')

    # # can't do batch correction with only one donor
    # with pytest.raises(ValueError):
    #     correct.normalize_expression(micro[0], norm='batch')


def test_remove_distance(donor_expression, atlas):
    expr = pd.concat(donor_expression).groupby('label').aggregate(np.mean)
    expr = expr.dropna(axis=1, how='any')
    coexpr = np.corrcoef(expr)
    for atlas_info in [None, atlas['info']]:
        out = correct.remove_distance(coexpr, atlas['image'], atlas_info)
        assert np.allclose(out, out.T)
        assert isinstance(out, np.ndarray)

    # subset expression data + and atlas_info
    coexpr = np.corrcoef(expr.iloc[:-1])
    removed_label = pd.read_csv(atlas_info).iloc[:-1]
    out = correct.remove_distance(coexpr, atlas['image'], removed_label,
                                  labels=removed_label.id)
    assert np.allclose(out, out.T)
    assert isinstance(out, np.ndarray)
    assert len(out) == len(removed_label)

    with pytest.raises(ValueError):
        correct.remove_distance(np.corrcoef(expr), atlas['image'],
                                removed_label, labels=removed_label.id)

    with pytest.raises(ValueError):
        correct.remove_distance(expr, atlas['image'], atlas['info'])


def test_resid_dist():
    dv = np.array([1, 2, 3, 4, 5])
    # residualizing against self should yield 0
    assert np.allclose(correct._resid_dist(dv, iv=dv), 0)
    # residualizing against perfectly anticorrelated should also yield 0
    assert np.allclose(correct._resid_dist(dv, iv=dv[::-1]), 0)
    # residualizing against scaled self should also yield 0 (intercept incl)
    assert np.allclose(correct._resid_dist(dv, iv=(dv + 10)), 0)
    # residualizing against constant should yield de-meaned input
    assert np.allclose(correct._resid_dist(dv, iv=np.ones_like(dv)),
                       dv - dv.mean())


def test_keep_stable_genes(donor_expression):
    for thr, per, rank in itertools.product(np.arange(0, 1, 0.1),
                                            [True, False],
                                            [True, False]):
        out = correct.keep_stable_genes(donor_expression, threshold=thr,
                                        percentile=per, rank=rank)
        assert all([isinstance(f, pd.DataFrame) for f in out])
        for df1, df2 in itertools.combinations(out, 2):
            assert df1.shape == df2.shape

    # check that `return_stability` provides expression and stability
    out, stab = correct.keep_stable_genes(donor_expression, threshold=0,
                                          return_stability=True)
    assert len(stab) == len(out[0].columns)
    assert np.all(out[0].columns == donor_expression[0].columns)



# -*- coding: utf-8 -*-
"""
Fixtures for all abagen tests
"""

import os
import pytest

from abagen.datasets import (fetch_desikan_killiany,
                             fetch_microarray,
                             fetch_rnaseq)


@pytest.fixture(scope='session')
def datadir(tmp_path_factory):
    dd = os.environ.get('ABAGEN_DATA')
    if dd is None:
        dd = str(tmp_path_factory.mktemp('abagen-data'))
    return dd


@pytest.fixture(scope='session')
def testfiles(datadir):
    return fetch_microarray(data_dir=datadir, donors=['12876', '15496'],
                            n_proc=2)


@pytest.fixture(scope='session')
def rnafiles(datadir):
    return fetch_rnaseq(data_dir=datadir, donors=['9861'])


@pytest.fixture(scope='session')
def atlas():
    return fetch_desikan_killiany()



# -*- coding: utf-8 -*-
"""
Defines information about the abagen distribution
"""

long_description = """
abagen: A toolbox for the Allen Brain Atlas genetics data
=========================================================

This package provides a Python interface for fetching and working with the
`Allen Human Brain Atlas`_ (AHBA) microarray expression data.

.. image:: https://travis-ci.org/rmarkello/abagen.svg?branch=master
   :target: https://travis-ci.org/rmarkello/abagen
.. image:: https://codecov.io/gh/rmarkello/abagen/branch/master/graph/badge.svg
   :target: https://codecov.io/gh/rmarkello/abagen
.. image:: https://readthedocs.org/projects/abagen/badge/?version=latest
   :target: https://abagen.readthedocs.io/en/stable
.. image:: https://img.shields.io/badge/License-BSD%203--Clause-blue.svg
   :target: https://opensource.org/licenses/BSD-3-Clause
.. image:: https://zenodo.org/badge/DOI/10.5281/zenodo.3451463.svg
   :target: https://doi.org/10.5281/zenodo.3451463

.. _readme_overview:

Overview
--------

In 2013, the Allen Institute for Brain Science released the `Allen Human Brain
Atlas`_, a dataset containing microarray expression data collected from six
human brains (Hawrylycz et al., 2012) . This dataset has offered an
unprecedented opportunity to examine the genetic underpinnings of the human
brain, and has already yielded novel insight into e.g., `adolescent brain
development <https://www.pnas.org/content/113/32/9105.long>`__ and `functional
brain organization <https://science.sciencemag.org/content/348/6240/
1241.long>`__.

However, in order to be effectively used in most analyses, the AHBA microarray
expression data often needs to be (1) collapsed into regions of interest (e.g.,
parcels or networks), and (2) combined across donors. While this may
potentially seem trivial, there are a number of analytic choices in these steps
that can dramatically influence the resulting data and any downstream analyses.
Arnatkevičiūte et al., 2019 provided a thorough treatment of this in a `recent
manuscript <https://www.sciencedirect.com/science/article/pii/
S1053811919300114>`__, demonstrating how the techniques and code used to
prepare the raw AHBA data have varied widely across published reports.

The current Python package, ``abagen``, aims to provide reproducible workflows
for processing and preparing the AHBA microarray expression data for analysis.

.. _readme_requirements:

Installation requirements
-------------------------
Currently, ``abagen`` works with Python 3.6+ and requires a few dependencies:

    - nibabel
    - numpy (>=1.14.0)
    - pandas, and
    - scipy

There are some additional (optional) dependencies you can install to speed up
some functions:

    - fastparquet, and
    - python-snappy

These latter packages are primarily used to facilitate loading the (rather
large!) microarray expression dataframes provided by the Allen Institute,

For detailed information on how to install ``abagen``, including these
dependencies, refer to our `installation instructions`_.

.. _readme_usage:

Quickstart
----------

At it's core, using ``abagen`` is as simple as:

.. code-block:: python

    >>> import abagen
    >>> expression = abagen.get_expression_data('myatlas.nii.gz')  # doctest: +SKIP

where ``'myatlas.nii.gz'`` points to a Nifti file that defines a brain
parcellation in MNI space.

This function can also be called from the command line with:

.. code-block:: bash

    $ abagen --output-file expression.csv myatlas.nii.gz

For more detailed instructions on how to use ``abagen`` please refer to our
`user guide`_!

.. _readme_development:

Development and getting involved
--------------------------------

If you've found a bug, are experiencing a problem, or have a question about
using the package, please head on over to our `GitHub issues`_ and make a new
issue with some information about it! Someone will try and get back to you
as quickly as possible, though please note that the primary developer for
``abagen`` (@rmarkello) is a graduate student so responses make take some time!

If you're interested in getting involved in the project: welcome |sparkles|!
We're thrilled to welcome new contributors. You should start by reading our
`code of conduct`_; all activity on ``abagen`` should adhere to the CoC. After
that, take a look at our `contributing guidelines`_ so you're familiar with the
processes we (generally) try to follow when making changes to the repository!
Once you're ready to jump in head on over to our issues to see if there's
anything you might like to work on.

.. _readme_acknowledgments:

Citing ``abagen``
-----------------

For up-to-date instructions on how to cite abagen please refer to our
`documentation <https://abagen.readthedocs.io/en/stable/citing.html>`_.

.. _readme_licensing:

License Information
-------------------

This codebase is licensed under the `3-clause BSD license`_. The full license
can be found in the `LICENSE`_ file in the ``abagen`` distribution.

Reannotated gene information located at ``abagen/data/reannotated.csv.gz`` and
individualized donor parcellations for the Desikan-Killiany atlas located at
``abagen/data/native_dk`` are taken from Arnatkevičiūte et al., 2018 and are
separately licensed under the `CC BY 4.0`_; these data can also be found on
`figshare <https://figshare.com/s/441295fe494375aa0c13>`__. Individualized
parcellations have been renumbered to match the MNI-space atlas originally
shipped with ``abagen``.

Corrected MNI coordinates used to match AHBA tissues samples to MNI space
located at ``abagen/data/corrected_mni_coordinates.csv`` are taken from the
`alleninf package`_, provided under the `3-clause BSD license`_.

All microarray expression data is copyrighted under `non-commercial reuse
policies`_ by the Allen Institute for Brain Science (© 2010 Allen Institute for
Brain Science. Allen Human Brain Atlas. Available from: `Allen Human Brain
Atlas`_).

All trademarks referenced herein are property of their respective holders.

.. |sparkles| replace:: ✨
.. |warning| replace:: 🚨
.. _3-clause BSD license: https://opensource.org/licenses/BSD-3-Clause
.. _Allen Human Brain Atlas: https://human.brain-map.org/
.. _alleninf package: https://github.com/chrisfilo/alleninf
.. _CC BY 4.0: https://creativecommons.org/licenses/by/4.0/legalcode
.. _code of conduct: https://github.com/rmarkello/abagen/blob/master/CODE_OF_CONDUCT.md
.. _contributing guidelines: https://github.com/rmarkello/abagen/blob/master/CONTRIBUTING.md
.. _contributors: https://github.com/rmarkello/abagen/graphs/contributors
.. _user guide: https://abagen.readthedocs.io/en/stable/usage.html
.. _GitHub issues: https://github.com/rmarkello/abagen/issues
.. _installation instructions: https://abagen.readthedocs.io/en/stable/installation.html
.. _LICENSE: https://github.com/rmarkello/abagen/blob/master/LICENSE
.. _non-commercial reuse policies: https://alleninstitute.org/legal/terms-use/
"""  # noqa



# -*- coding: utf-8 -*-

import random
import numpy as np
import pytest
from abagen.mouse import io, structure

RANDOM = random.Random(1234)
STRUCTS = io.fetch_allenref_structures()
COORD_ID = {
    'sagittal': (7800, 3400, 1050),
    'coronal': (7800, 3400, 10350)
}
COORD_ACRONYM = {
    'sagittal': (5740, 2060, 2730),
    'coronal': (5740, 2060, 8670)
}


def test_get_structure_info():
    samples = STRUCTS.loc[RANDOM.sample(range(len(STRUCTS)), 10)]

    acronyms = structure.get_structure_info(id=samples['id'],
                                            attributes='acronym')
    assert sorted(acronyms['acronym']) == sorted(samples['acronym'])

    names = structure.get_structure_info(id=samples['id'],
                                         attributes='name')
    assert sorted(names['name']) == sorted(samples['name'])

    with pytest.raises(ValueError):
        structure.get_structure_info(acronym=samples['acronym'],
                                     attributes='invalid')

    # multiple attributes
    info = structure.get_structure_info(id=samples['id'],
                                        attributes=['name', 'acronym'])
    assert sorted(acronyms['acronym']) == sorted(samples['acronym'])
    assert sorted(info['name']) == sorted(samples['name'])

    # all attributes
    info = structure.get_structure_info(id=samples['id'])
    assert len(info.columns) == 16

    # exception: invalid gene identifiers
    with pytest.raises(ValueError):
        structure.get_structure_info(id=-100000)
    with pytest.raises(ValueError):
        structure.get_structure_info(acronym='notastructure')


@pytest.mark.parametrize('space', ['coronal', 'sagittal'])
def test_get_structure_coordinates(space):
    info = structure.get_structure_coordinates(id=1018,
                                               reference_space=space)
    assert tuple(np.asarray(info[['x', 'y', 'z']])[0]) == COORD_ID[space]

    info = structure.get_structure_coordinates(acronym='SSp',
                                               reference_space=space)
    assert tuple(np.asarray(info[['x', 'y', 'z']])[0]) == COORD_ACRONYM[space]

    # exception: structure is invalid
    with pytest.raises(ValueError):
        structure.get_structure_coordinates(id=-1000000)
    with pytest.raises(ValueError):
        structure.get_structure_coordinates(acronym='notastructure')



# -*- coding: utf-8 -*-
"""
Tests for abagen.allen module
"""

import nibabel as nib
import numpy as np
import pandas as pd
import pytest

from abagen import allen
from abagen.utils import check_img


def test_vanilla_get_expression_data(testfiles, atlas):
    out = allen.get_expression_data(atlas['image'], donors=['12876', '15496'])
    assert np.allclose(out.index, range(1, 84))
    assert out.index.name == 'label'
    assert out.columns.name == 'gene_symbol'


@pytest.mark.parametrize('opts', [
    ({'atlas_info': True}),
    ({'exact': False}),
    ({'reannotated': False}),
    ({'atlas_info': True, 'exact': False}),
])
def test_extra_get_expression_data(testfiles, atlas, opts):
    if opts.get('atlas_info'):
        opts['atlas_info'] = atlas['info']

    out = allen.get_expression_data(atlas['image'], donors=['12876', '15496'],
                                    **opts)
    assert out.index.name == 'label'
    assert out.columns.name == 'gene_symbol'


def test_individualized_get_expression_data(testfiles):
    atlas = allen.datasets.fetch_desikan_killiany(native=True)
    out = allen.get_expression_data(atlas['image'], donors=['12876', '15496'])
    assert out.index.name == 'label'
    assert out.columns.name == 'gene_symbol'


def test_get_expression_data_errors(testfiles, atlas):
    # invalid probe_selection method
    with pytest.raises(ValueError):
        allen.get_expression_data(atlas['image'], donors=['12876', '15496'],
                                  probe_selection='nonsense')

    # cannot use diff_stability with only one donor
    with pytest.raises(ValueError):
        allen.get_expression_data(atlas['image'], donors=['12876'],
                                  probe_selection='diff_stability')


def test_missing_labels(testfiles, atlas):
    # remove some labels from atlas image so numbers are non-sequential
    remove = [10, 20, 60]

    # subset atlas image
    img = check_img(atlas['image'])
    img_data = np.asarray(img.dataobj)
    for i in remove:
        img_data[img_data == i] = 0
    img = img.__class__(img_data, img.affine)

    # subset atlas info
    info = pd.read_csv(atlas['info'])
    info = info[~info['id'].isin(remove)]

    # test get expression
    out, counts = allen.get_expression_data(img, info,
                                            exact=False, return_counts=True,
                                            donors=['12876', '15496'])
    assert out.index.name == 'label'
    assert out.columns.name == 'gene_symbol'
    assert len(out) == len(info)

    assert isinstance(counts, pd.DataFrame)
    assert counts.shape == (len(info), len(testfiles))


def test_get_samples_in_mask(testfiles, atlas):
    allexp, allcoords = allen.get_samples_in_mask(donors=['12876', '15496'])
    cortexp, cortcoords = allen.get_samples_in_mask(mask=atlas['image'],
                                                    donors=['12876', '15496'])

    # exp + coords shape as expected?
    assert len(allexp) == len(allcoords) and len(cortexp) == len(cortcoords)
    assert allcoords.shape[-1] == 3 and cortcoords.shape[-1] == 3
    for df in [allexp, cortexp]:
        assert df.index.name == 'well_id'
        assert df.columns.name == 'gene_symbol'

    # providing the cortical mask (atlas) should reduce # of samples returned
    assert len(allexp) > len(cortexp)


def test_coerce_atlas_to_dict(testfiles, atlas):
    img, info = atlas['image'], atlas['info']
    donors = ['12876', '15496']

    # test providing single atlas file
    atl, inf, same = allen.coerce_atlas_to_dict(img, donors, info)
    assert same
    assert sorted(atl.keys()) == sorted(donors)
    assert isinstance(inf, pd.DataFrame)
    imgs = list(atl.values())
    assert all(imgs[0] is a for a in imgs[1:])

    # test providing pre-loaded atlas file
    atl, inf, same = allen.coerce_atlas_to_dict(nib.load(img), donors, info)
    assert same
    assert sorted(atl.keys()) == sorted(donors)
    assert isinstance(inf, pd.DataFrame)
    imgs = list(atl.values())
    assert all(imgs[0] is a for a in imgs[1:])

    # test providing dictionary for atlas
    atlas_dict = {d: img for d in donors}
    atl, inf, same = allen.coerce_atlas_to_dict(atlas_dict, donors)
    assert not same
    assert sorted(atl.keys()) == sorted(donors)
    assert inf is None
    imgs = list(atl.values())
    assert not any(imgs[0] is a for a in imgs[1:])

    with pytest.raises(ValueError):
        allen.coerce_atlas_to_dict(atlas_dict, donors + ['9861'])

    with pytest.raises(TypeError):
        allen.coerce_atlas_to_dict('notanatlas', donors)



# -*- coding: utf-8 -*-
"""
Functions to fetch (and load) mouse gene and structure lists from Allen API
"""

import os.path as op
from pkg_resources import resource_filename

import pandas as pd

from .utils import _make_api_query


def fetch_allenref_genes(entry_type=None, cache=True, verbose=True):
    """
    Loads all genes from Allen Reference database

    Parameters
    ----------
    entry_type: {'id', 'acronym', 'name'}, optional
        The type of gene identifier to load. Specifying 'id' returns a list of
        numerical gene IDs, 'acronym' returns a list of short-form gene
        acronyms, and 'name' returns full gene names. If not specified, returns
        a dataframe with all information. Default: None
    cache : bool, optional
        Whether to use cached gene information (if it exists). Setting to False
        will overwrite cache. Default: True
    verbose : bool, optional
        Whether to print status message. Default: True

    Returns
    -------
    genes : list or :obj:`pandas.DataFrame`
        Genes in Allen Reference database

    Notes
    -----
    May require internet access to make query to the Allen API (which will take
    some time); after query is made once the results are cached.
    """

    # check that provided entry_type is valid
    entries = ['id', 'acronym', 'name']
    if entry_type is not None and entry_type not in entries:
        raise ValueError('Provided entry_type {} is not valid. Specified '
                         'entry_type must be one of {}.'
                         .format(entry_type, entries))

    # if file doesn't exist or we want to overwrite the cache for some reason
    # download the data from the Allen API
    fname = resource_filename('abagen', 'data/allen_reference_genes.csv')
    if not op.isfile(fname) or not cache:
        if verbose:
            print('Gene information not available locally; querying '
                  'Allen API for information. This may take some time...')
        out = _make_api_query('Gene', criteria='products[id$eq1]',
                              attributes=entries, suffix='num_rows=all')
        genes = pd.DataFrame(out)
        # sort entries by gene ID
        genes = genes.sort_values('id').reset_index(drop=True)
        # save information to disk
        genes.to_csv(fname, index=False)
    else:
        genes = pd.read_csv(fname)[entries]

    # extract only relevant entry_type, if desired
    if entry_type is not None:
        genes = genes[entry_type].tolist()

    return genes


def fetch_allenref_structures(entry_type=None, cache=True, verbose=True):
    """
    Loads all anatomical structures in the Allen Reference Atlas

    Parameters
    ----------
    entry_type: {'id', 'acronym', 'name'}, optional
        The type of structural identifier to load. Specifying 'id' returns a
        list of numerical structure IDs, 'acronym' returns a list of short-form
        structure acronyms, and 'name' returns full structure names. If not
        specified, returns a dataframe with all information. Default: None
    cache : bool, optional
        Whether to use cached structure information (if it exists). Setting to
        False will overwrite cache. Default: True
    verbose : bool, optional
        Whether to print status message. Default: True

    Returns
    -------
    structures : list or :obj:`pandas.DataFrame`
        Anatomical structures in Allen Reference Atlas

    Notes
    -----
    May require internet access to make query to the Allen API (which will take
    some time); after query is made once the results are cached.
    """

    entries = ['id', 'acronym', 'name']
    if entry_type is not None and entry_type not in entries:
        raise ValueError('Provided entry_type {} is not valid. Specified '
                         'entry_type must be one of {}.'
                         .format(entry_type, entries))

    fname = resource_filename('abagen', 'data/allen_reference_atlas.csv')
    if not op.isfile(fname) or not cache:
        if verbose:
            print('Structure information not available locally; querying '
                  'Allen API for information...')
        out = _make_api_query('Structure', criteria='ontology[id$eq1]',
                              attributes=entries, suffix='num_rows=all')
        structures = pd.DataFrame(out)
        # sort entries by structure ID
        structures = structures.sort_values('id').reset_index(drop=True)
        # save information to disk
        structures.to_csv(fname, index=False)
    else:
        structures = pd.read_csv(fname)[entries]

    # extract only relevant entry_type, if desired
    if entry_type is not None:
        structures = structures[entry_type].tolist()

    return structures


def fetch_rubinov2015_structures(entry_type=None):
    """
    Loads subset of anatomical structures in Allen Reference Atlas used in [1]_

    Parameters
    ----------
    entry_type: {'id', 'acronym', 'name'}, optional
        The type of structural identifier to load. Specifying 'id' returns a
        list of numerical structure IDs, 'acronym' returns a list of short-form
        structure acronyms, and 'name' returns full structure names. If not
        specified, returns a dataframe with all information. Default: None

    Returns
    -------
    structures : list or :obj:`pandas.DataFrame`
        Anatomical structures in Allen Reference Atlas from [1]_

    References
    ----------
    .. [1] Rubinov, M., Ypma, R. J., Watson, C., & Bullmore, E. T. (2015).
       Wiring cost and topological participation of the mouse brain connectome.
       Proceedings of the National Academy of Sciences, 112(32), 10032-10037.
    """

    entries = ['id', 'acronym', 'name']
    if entry_type is not None and entry_type not in entries:
        raise ValueError('Provided entry_type {} is not valid. Specified '
                         'entry_type must be one of {}.'
                         .format(entry_type, entries))

    fname = resource_filename('abagen', 'data/rubinov2015_pnas.csv')
    structures = pd.read_csv(fname)[entries]

    if entry_type is not None:
        structures = structures[entry_type].tolist()

    return structures



# -*- coding: utf-8 -*-
"""
Functions for processing and correcting gene expression data
"""

import functools
import itertools
import warnings

import numpy as np
import pandas as pd
import scipy.stats as sstats
from scipy.spatial.distance import cdist

from . import utils


def _unpack_tuple(var):
    """
    Returns first entry of var if there is only one entry

    Parameters
    ----------
    var : tuple

    Returns
    -------
    var
    """

    if len(var) == 1:
        return var[0]
    return var


def _batch_correct(data):
    """
    Performs batch correction on `data`

    Parameters
    ----------
    data : list of (S, G) pandas.DataFrame
        Microarray expression data, where `S` is samples (or regions) and `G`
        is genes

    Returns
    -------
    residualized : list of (S, G) pandas.DataFrame
        Batch-corrected microarray expression data
    """

    if len(data) < 2:
        raise ValueError('Cannot perform batch correction with one batch.')

    # need to drop NaNs or the lstsq fit will choke
    data = [np.asarray(f) for f in data]
    data = [f[np.logical_not(np.all(np.isnan(f), axis=1))] for f in data]

    # generate donor label / "batch" array
    n_samp = [len(d) for d in data]
    batch = np.repeat(range(len(data)), n_samp)
    batch = np.column_stack([batch == f for f in np.unique(batch)]).astype(int)
    # add intercept for fit (but we won't regress this out)
    batch = np.column_stack([batch, np.ones(len(batch), dtype=int)])

    # fit least squares and residualize
    raw = np.row_stack(data)
    betas = np.linalg.lstsq(batch, raw, rcond=None)[0]
    resid = raw - (batch[:, :-1] @ betas[:-1])
    residualized = np.split(resid, np.cumsum(n_samp)[:-1])

    return residualized


def _rescale(data, low=0, high=1, axis=0):
    """
    Rescales `data` to range [`low`, `high`]

    Parameters
    ----------
    data : array_like
        Input data array to be rescaled
    low : float, optional
        Lower bound for rescaling. Default: 0
    high : float, optional
        Upper bound for rescaling. Default: 1
    axis : int, optional
        Axis of `data` to scale along. Default: 0

    Returns
    -------
    rescaled : np.ndarray
        Rescaled data
    """

    data = np.asanyarray(data)

    dmin = data.min(axis=axis, keepdims=True)
    dmax = data.max(axis=axis, keepdims=True)
    rescaled = ((data - dmin) / (dmax - dmin)) * (high - low) + low

    return rescaled


def _sigmoid(data, axis=0):
    """
    Normalizes `data` with a standard sigmoid function

    Parameters
    ----------
    data : array_like
        Input data array to be transformed
    axis : int, optional
        Axis of `data` to be normalized

    Returns
    -------
    normed : array_like
        Normalized input `data`
    """

    data = np.asanyarray(data)

    mean = np.mean(data, axis=axis, keepdims=True)
    std = np.std(data, ddof=1, axis=axis, keepdims=True)
    normed = 1 / (1 + np.exp(-(np.asarray(data) - mean) / std))

    return normed


def _scaledsig(data, low=0, high=1, axis=0):
    """
    Normalizes `data` with a scaled sigmoid function

    Parameters
    ----------
    data : array_like
        Input data array to be transformed
    low : float, optional
        Lower bound for rescaling. Default: 0
    high : float, optional
        Upper bound for rescaling. Default: 1
    axis : int, optional
        Axis of `data` to be normalized

    Returns
    -------
    normed : array_like
        Normalized input `data`
    """

    sig = _sigmoid(np.asanyarray(data), axis=axis)  # sigmoid transform
    normed = _rescale(sig, low=low, high=high, axis=axis)  # scales data

    return normed


def _rs(data, axis=0):
    """
    Normalizes `data` with a robust sigmoid function

    Parameters
    ----------
    data : array_like
        Input data array to be transformed
    axis : int, optional
        Axis of `data` to be normalized

    Returns
    -------
    normed : array_like
        Normalized input `data`
    """

    data = np.asanyarray(data)

    # calculate sigmoid normalization
    med = np.median(data, axis=axis, keepdims=True)
    iqr = sstats.iqr(data, axis=axis, scale='normal', keepdims=True)
    normed = 1 / (1 + np.exp(-(data - med) / iqr))

    return normed


def _srs(data, low=0, high=1, axis=0):
    """
    Normalizes `data` with a scaled robust sigmoid function

    Parameters
    ----------
    data : array_like
        Input data array to be transformed
    low : float, optional
        Lower bound for rescaling. Default: 0
    high : float, optional
        Upper bound for rescaling. Default: 1
    axis : int, optional
        Axis of `data` to be normalized

    Returns
    -------
    normed : array_like
        Normalized input `data`
    """

    srs = _rs(np.asanyarray(data), axis=axis)  # robust sigmoid transform
    normed = _rescale(srs, low=low, high=high, axis=axis)  # scales data

    return normed


def _scaledsig_qnt(data, low=0, high=1, axis=0):
    """
    Normalizes `data` with capped-quantile scaled sigmoid function

    Caps `data` at 5th and 95th percentiles and then uses `_scaledsig()`

    Parameters
    ----------
    data : array_like
        Input data array to be transformed
    low : float, optional
        Lower bound for rescaling. Default: 0
    high : float, optional
        Upper bound for rescaling. Default: 1
    axis : int, optional
        Axis of `data` to be normalized

    Returns
    -------
    normed : array_like
        Normalized input `data`
    """

    data = np.asanyarray(data)
    lq, hq = np.percentile(data, [5, 95], axis=axis, keepdims=True)
    mdata = np.ma.masked_array(data, np.logical_or(data > hq, data < lq))
    normed = np.asarray(_scaledsig(mdata, low=low, high=high, axis=axis))

    return normed


def _mixedsig(data, low=0, high=1, axis=0):
    """
    Uses `_scaledsig()` if IQR is 0; otherwise, uses `_srs()`

    Parameters
    ----------
    data : array_like
        Input data array to be transformed
    low : float, optional
        Lower bound for rescaling. Default: 0
    high : float, optional
        Upper bound for rescaling. Default: 1
    axis : int, optional
        Axis of `data` to be normalized

    Returns
    -------
    normed : array_like
        Normalized input `data`
    """

    data = np.asanyarray(data)

    iqr = sstats.iqr(data, axis=axis, scale='normal')
    mask = iqr == 0

    normed = np.zeros_like(data)
    normed[:, mask] = _scaledsig(data[:, mask])
    normed[:, ~mask] = _srs(data[:, ~mask])

    # constant columns set to 0
    return np.nan_to_num(normed)


NORMALIZATION_METHODS = dict(
    sig=functools.partial(_sigmoid, axis=0),
    scaled_sig=functools.partial(_scaledsig, low=0, high=1, axis=0),
    scaled_sig_qnt=functools.partial(_scaledsig_qnt, low=0, high=1, axis=0),
    mixed_sig=functools.partial(_mixedsig, low=0, high=1, axis=0),
    rs=functools.partial(_rs, axis=0),
    srs=functools.partial(_srs, low=0, high=1, axis=0),
    center=lambda x: x - x.mean(axis=0, keepdims=True),
    zscore=functools.partial(sstats.zscore, axis=0, ddof=1),
    minmax=functools.partial(_rescale, low=0, high=1, axis=0)
)
# aliases
for alias, orig in [('demean', 'center'),
                    ('rsig', 'rs'),
                    ('robust_sigmoid', 'rs'),
                    ('scaled_rsig', 'srs'),
                    ('scaled_robust_sigmoid', 'srs'),
                    ('sigmoid', 'sig'),
                    ('scaled_sigmoid', 'scaled_sig'),
                    ('scaled_sigmoid_quantiles', 'scaled_sig_qnt'),
                    ('mixed_sigmoid', 'mixed_sig')]:
    NORMALIZATION_METHODS[alias] = NORMALIZATION_METHODS[orig]


def normalize_expression(expression, norm='srs', ignore_warn=False):
    """
    Performs normalization on `expression` data

    Parameters
    ----------
    expression : list of (S, G) pandas.DataFrame
        Microarray expression data, where `S` is samples (or regions) and `G`
        is genes
    norm : str, optional
        Function with which to normalize expression data. See Notes for more
        information on options. Default: 'scaled_robust_sigmoid'
    ignore_warn : bool, optional
        Whether to suppress potential warnings raised by normalization.
        Default: False

    Returns
    -------
    normalized : list of (S, G) pandas.DataFrame
        Data from `expression` normalized separately for each gene

    Notes
    -----
    The following methods can be used for normalizing gene expression values
    for each donor (adapted from [PC2]_):

    1. ``norm='center'``

    Removes the mean of data in each column. Aliased to 'demean'

    2. ``norm='zscore'``

    Applies a basic z-score (subtract mean, divide by standard deviation) to
    each column; uses degrees of freedom equal to one for standard deviation

    3. ``norm='minmax'``

    Scales data in each column to the unit normal (i.e., range 0-1)

    4. ``norm='sigmoid'``

    Applies a sigmoidal transform function to normalize data in each column.
    Aliased to 'sig'

    5. ``norm='scaled_sigmoid'``

    Combines 'sigmoid' and 'minmax'. Aliased to 'scaled_sig'

    6. ``norm='scaled_sigmoid_quantiles'``

    Caps input data at the 5th and 95th percentiles before performing the
    'scaled_sigmoid' transform. Aliased to 'scaled_sig_qnt'

    7. ``norm='robust_sigmoid'``

    Uses a robust sigmoid function ([PC1]_) to normalize data in each column.
    Aliased to 'rs' and 'rsig'

    8. ``norm='scaled_robust_sigmoid'``

    Combines 'robust_sigmoid' and 'minmax'. Alised to 'srs' and 'scaled_rsig'

    9. ``norm='mixed_sigmoid'``

    Uses 'scaled_sigmoid' transform for columns where the IQR is 0; otherwise,
    uses the 'scaled_robust_sigmoid' transform. Aliased to 'mixed_sig'

    10. ``norm='batch'``

    Uses a linear model to remove donor effects from data. Differs from other
    methods in that all donors are simultaneously fit to the same model and
    data are residualized based on estimated betas. Linear model includes the
    intercept but it is not removed during residualization

    References
    ----------
    .. [PC1] Fulcher, B. D., & Fornito, A. (2016). A transcriptional signature
       of hub connectivity in the mouse connectome. Proceedings of the National
       Academy of Sciences, 113(5), 1435-1440.
    .. [PC2] Fulcher, B. D., Little, M. A., & Jones, N. S. (2013). Highly
       comparative time-series analysis: the empirical structure of time series
       and their methods. Journal of the Royal Society Interface, 10(83),
       20130048
    """

    try:
        normfunc = NORMALIZATION_METHODS[norm]
    except KeyError:
        raise ValueError('Provided value for `norm` not recognized. Must be '
                         f'one of {list(NORMALIZATION_METHODS)}. Received: '
                         f'{norm}')

    # FIXME: I hate having to do this...
    if isinstance(expression, pd.DataFrame):
        expression = [expression]

    if norm == 'batch':
        corrected = _batch_correct(expression)

    normexp = []
    for n, exp in enumerate(expression):
        # get non-NaN values
        notna = np.logical_not(exp.isna().all(axis=1))
        data = np.asarray(exp)[notna]

        kwargs = dict(all='ignore') if ignore_warn else {}
        with np.errstate(**kwargs):
            # normalize the data (however was specified)
            normed = normfunc(data) if norm != 'batch' else corrected[n]

        # recreate dataframe and fill non-NaN values
        normalized = pd.DataFrame(np.nan, columns=exp.columns, index=exp.index)
        normalized.loc[notna] = normed
        normexp.append(normalized)

    return _unpack_tuple(normexp)


def remove_distance(coexpression, atlas, atlas_info=None, labels=None):
    """
    Corrects for distance-dependent correlation effects in `coexpression`

    Regresses Euclidean distance between regions in `atlas` from correlated
    gene expression array `coexpression`. If `atlas_info` is provided different
    connection types (e.g., cortex-cortex, cortex-subcortex, subcortex-
    subcortex) will be residualized independently.

    Parameters
    ----------
    coexpression : (R x R) array_like
        Correlated gene expression array, where `R` is the number of regions,
        as generated with e.g., `numpy.corrcoef(expression)`.
    atlas : niimg-like object
        A parcellation image in MNI space, where each parcel is identified by a
        unique integer ID
    atlas_info : str or pandas.DataFrame, optional
        Filepath to or pre-loaded dataframe containing information about
        `atlas`. Must have at least columns 'id', 'hemisphere', and 'structure'
        containing information mapping atlas IDs to hemisphere (i.e, "L", "R")
        and broad structural class (i.e., "cortex", "subcortex", "cerebellum").
        Default: None
    labels : (N,) array_like, optional
        If only a subset `N` of the ROIs in `atlas` were used to generate the
        `coexpression` array this array should specify which to consider. Not
        specifying this may cause a ValueError if `atlas` and `atlas_info` do
        not match. Default: None

    Returns
    -------
    residualized : (R, R) numpy.ndarray
        Provided `coexpression` data residualized against spatial distance
        between region pairs
    """

    # load atlas_info, if provided
    atlas = utils.check_img(atlas)
    if atlas_info is not None:
        atlas_info = utils.check_atlas_info(atlas, atlas_info, labels=labels)
        if labels is not None and len(labels) != len(coexpression):
            raise ValueError(f'Provided labels {labels} are a different '
                             'length than provided coexpression matrix of '
                             f'size {coexpression.shape}. Please confirm '
                             'inputs and try again.')

    # check that provided coexpression array is symmetric
    if not np.allclose(coexpression, coexpression.T, atol=1e-10):
        raise ValueError('Provided coexpression matrix is not symmetric')

    # we'll do basic Euclidean distance correction for now
    # TODO: implement gray matter volume / cortical surface path distance
    centroids = utils.get_centroids(atlas, labels=labels)
    dist = cdist(centroids, centroids, metric='euclidean')

    corr_resid = np.zeros_like(coexpression)
    triu_inds = np.triu_indices_from(coexpression, k=1)
    # if no atlas_info, just residualize all correlations against distance
    if atlas_info is None:
        corr_resid[triu_inds] = _resid_dist(coexpression[triu_inds],
                                            dist[triu_inds])
    # otherwise, we can residualize the different connection types separately
    else:
        triu_inds = np.ravel_multi_index(triu_inds, corr_resid.shape)
        coexpression, dist = coexpression.ravel(), dist.ravel()
        types = ['cortex', 'subcortex']
        for src, tar in itertools.combinations_with_replacement(types, 2):
            # get indices of sources and targets
            sources = np.where(atlas_info.structure == src)[0]
            targets = np.where(atlas_info.structure == tar)[0]
            inds = np.ravel_multi_index(np.ix_(sources, targets),
                                        corr_resid.shape)
            if src != tar:  # e.g., cortex + subcortex
                rev = np.ravel_multi_index(np.ix_(targets, sources),
                                           corr_resid.shape)
                inds = np.append(inds.ravel(), rev.ravel())
            # find intersection of source / target indices + upper triangle
            inds = np.intersect1d(triu_inds, inds)
            back = np.unravel_index(inds, corr_resid.shape)
            # residualize
            corr_resid[back] = _resid_dist(coexpression[inds], dist[inds])

    corr_resid = (corr_resid + corr_resid.T + np.eye(len(corr_resid)))

    return corr_resid


def _resid_dist(dv, iv):
    """
    Calculates residuals of `dv` after controlling for `iv`

    Parameters
    ----------
    dv : array_like
        Dependent variable
    iv : array_like
        Independent variable; removed from `dv`

    Returns
    -------
    residuals : array_like
        Residuals of `dv` after controlling for `iv`
    """
    dv = dv.squeeze()
    distance = np.column_stack((iv, np.ones_like(iv)))
    betas, *rest = np.linalg.lstsq(distance, dv, rcond=None)
    residuals = dv - (distance @ betas)

    return residuals.squeeze()


def keep_stable_genes(expression, threshold=0.9, percentile=True, rank=True,
                      return_stability=False):
    """
    Removes genes in `expression` with differential stability < `threshold`

    Calculates the similarity of gene expression across brain regions for every
    pair of donors in `expression`. Similarity is averaged across donor pairs
    and genes whose mean similarity falls below `threshold` are removed.

    Parameters
    ----------
    expression : list of (R, G) pandas.DataFrame
        Where each entry is the microarray expression of `R` regions across `G`
        genes for a given donor
    threshold : [0, 1] float, optional
        Minimum required average similarity (e.g, correlation) across donors
        for a gene to be retained. Default: 0.1
    percentile : bool, optional
        Whether to treat `threshold` as a percentile instead of an absolute
        cutoff. For example, `threshold=0.9` and `percentile=True` would
        retain only those genes with a differential stability in the top 10% of
        all genes, whereas `percentile=False` would retain only those genes
        with differential stability > 0.9. Default: True
    rank : bool, optional
        Whether to calculate similarity as Spearman correlation instead of
        Pearson correlation. Default: True
    return_stability : bool, optional
        Whether to return stability estimates for each gene in addition to
        expression data. Default: False

    Returns
    -------
    expression : list of (R, Gr) pandas.DataFrame
        Microarray expression for `R` regions across `Gr` genes, where `Gr` is
        the number of retained genes
    stability : (G,) numpy.ndarray
        Stability (average correlation) of each gene across pairs of donors.
        Only returned if ``return_stability=True``
    """

    # get number of donors and number of genes
    num_subj = len(expression)
    num_gene = expression[0].shape[-1]

    # rank data, if necessary
    for_corr = expression if not rank else [e.rank() for e in expression]

    # get correlation of gene expression across regions for all donor pairs
    gene_corrs = np.zeros((num_gene, sum(range(num_subj))))
    for n, (s1, s2) in enumerate(itertools.combinations(range(num_subj), 2)):
        regions = np.intersect1d(for_corr[s1].dropna(axis=0, how='all').index,
                                 for_corr[s2].dropna(axis=0, how='all').index)
        gene_corrs[:, n] = utils.efficient_corr(for_corr[s1].loc[regions],
                                                for_corr[s2].loc[regions])

    # average similarity across donors (ignore NaNs)
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=RuntimeWarning,
                                message='Mean of empty slice')
        gene_corrs = np.nan_to_num(np.nanmean(gene_corrs, axis=1))

    # calculate absolute threshold if percentile is desired
    if percentile:
        threshold = np.percentile(gene_corrs, threshold * 100)
    keep_genes = gene_corrs >= threshold
    expression = [e.iloc[:, keep_genes] for e in expression]

    if return_stability:
        return expression, gene_corrs

    return expression



# -*- coding: utf-8 -*-

import pandas as pd
import pytest

from abagen.mouse import io

entries = ['id', 'acronym', 'name']


def test_fetch_allenref_genes():
    # invalid entry_type fails
    with pytest.raises(ValueError):
        io.fetch_allenref_genes(entry_type='wrong_entry_type')

    # specifying entry_type returns list
    genes = io.fetch_allenref_genes(entry_type='acronym')
    assert isinstance(genes, list)
    assert all(isinstance(f, str) for f in genes)

    # no specification returns data frame
    genes = io.fetch_allenref_genes()
    assert isinstance(genes, pd.DataFrame)
    assert genes.shape == (19991, 3)
    assert list(genes.columns) == entries


def test_fetch_allenref_structures():
    with pytest.raises(ValueError):
        io.fetch_allenref_structures(entry_type='wrong_entry_type')

    structures = io.fetch_allenref_structures(entry_type='acronym')
    assert isinstance(structures, list)
    assert all(isinstance(f, str) for f in structures)

    structures = io.fetch_allenref_structures()
    assert isinstance(structures, pd.DataFrame)
    assert structures.shape == (1327, 3)
    assert list(structures.columns) == entries


def test_fetch_rubinov2015_structures():
    with pytest.raises(ValueError):
        io.fetch_rubinov2015_structures(entry_type='wrong_entry_type')

    structures = io.fetch_rubinov2015_structures(entry_type='acronym')
    assert isinstance(structures, list)
    assert all(isinstance(f, str) for f in structures)

    structures = io.fetch_rubinov2015_structures()
    assert isinstance(structures, pd.DataFrame)
    assert structures.shape == (56, 3)



# -*- coding: utf-8 -*-
"""
Tests for abagen.samples module
"""

import numpy as np
import pandas as pd
import pytest

from abagen import samples_
from abagen.utils import first_entry, flatten_dict

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#  generate fake data (based largely on real data) so we know what to expect  #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


@pytest.fixture(scope='module')
def ontology():
    """ Fake ontology dataframe
    """
    sid = [4251, 4260, 4322, 4323, 9422]
    hemi = ['L', 'R', 'L', 'R', np.nan]
    acronym = ['S', 'S', 'Cl', 'Cl', 'CC']
    path = [
        '/4005/4006/4007/4008/4219/4249/12896/4251/',
        '/4005/4006/4007/4008/4219/4249/12896/4260/',
        '/4005/4006/4007/4275/4321/4322/',
        '/4005/4006/4007/4275/4321/4323/',
        '/4005/9352/9418/9422/',
    ]
    name = [
        'subiculum, left',
        'subiculum, right',
        'claustrum, left',
        'claustrum, right',
        'central canal',
    ]
    return pd.DataFrame(dict(id=sid, hemisphere=hemi, name=name,
                             acronym=acronym, structure_id_path=path))


@pytest.fixture(scope='module')
def mm_annotation():
    """ Fake annotation dataframe with some samples mislabelled
    """
    mni_x = [-10, -20, 30, 40, 0]
    sid = [4251, 4323, 4323, 4251, 9422]
    sacr = ['S', 'Cl', 'Cl', 'S', 'CC']
    sname = [
        'subiculum, left',
        'claustrum, right',
        'claustrum, right',
        'subiculum, left',
        'central canal'
    ]
    ind = pd.Series(range(len(sid)), name='sample_id')
    return pd.DataFrame(dict(mni_x=mni_x, structure_id=sid,
                             structure_acronym=sacr, structure_name=sname),
                        index=ind)


@pytest.fixture(scope='module')
def annotation(mm_annotation):
    """ Fake annotation dataframe
    """
    out = mm_annotation.loc[[0, 2, 4]].reset_index(drop=True)
    out.index.name = 'sample_id'
    return out


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# test all the functions on our generated fake data so we know what to expect #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


def test_update_mni_coords():
    # xyz coordinates are getting replaced so who cares about the original
    # but ids are important and need to be real!
    x = y = z = [-10, 20, 30, 40]
    ids = [594, 2985, 1058, 1145]
    annotation = pd.DataFrame(dict(mni_x=x, mni_y=y, mni_z=z, well_id=ids))
    out = samples_.update_mni_coords(annotation)

    # confirm that no samples were lost / reordered during the update process
    # and that we still have all our columns
    assert np.all(out['well_id'] == annotation['well_id'])
    assert np.all(out.columns == annotation.columns)

    # but DO confirm that _something_ changes about the dataframes (i.e., our
    # bogus coordinates should be different)
    with pytest.raises(AssertionError):
        pd.testing.assert_frame_equal(out, annotation)
    assert np.all(out['mni_x'] != annotation['mni_x'])
    assert np.all(out['mni_y'] != annotation['mni_y'])
    assert np.all(out['mni_z'] != annotation['mni_z'])

    # if we provide invalid well_ids we should receive an error!
    annotation['well_id'] = [594, 2985, 1058, 99999999999]
    with pytest.raises(KeyError):
        samples_.update_mni_coords(annotation)


@pytest.mark.parametrize('path, expected', [
    ('/4005/4006/4007/4275/4276/4277/4278/12899/4286/', 'subcortex'),
    ('/4005/4006/4007/4275/4327/4341/4342/4344/', 'subcortex'),
    ('/4005/4006/4007/4008/4084/4103/4111/4112/4113/', 'cortex'),
    ('/4005/4006/4833/4696/4697/12930/12931/12933/4751/', 'cerebellum'),
    ('/4005/4006/9512/9676/9677/9680/9681/', 'brainstem'),
    ('/4005/4006/4833/9131/9132/9133/9492/', 'brainstem'),
    ('/4005/9218/9298/12959/265505622/', 'white matter'),
    ('/4005/9218/9219/9220/9227/', 'white matter'),
    ('/4005/9352/9418/9419/9708/', 'other'),
    ('/4005/9352/9353/9400/9402/', 'other'),
    ('/4005/4006/4833', None),
    ('thisisnotapath', None),  # TODO: should this error?
])
def test__get_struct(path, expected):
    out = samples_._get_struct(path)
    assert out == expected if expected is not None else out is None


def test_drop_mismatch_samples(mm_annotation, ontology):
    # here's what we expect (i.e., indices 1 & 3 are dropped and the structure
    # for the remaining samples is correctly extracted from the paths)
    expected = pd.DataFrame(dict(hemisphere=['L', 'R', np.nan],
                                 mni_x=[-10, 30, 0],
                                 structure_acronym=['S', 'Cl', 'CC'],
                                 structure=['subcortex', 'subcortex', 'other'],
                                 structure_id=[4251, 4323, 9422],
                                 structure_name=['subiculum, left',
                                                 'claustrum, right',
                                                 'central canal']),
                            index=pd.Series([0, 2, 4], name='sample_id'))

    # do we get what we expect? (ignore ordering of columns / index)
    out = samples_.drop_mismatch_samples(mm_annotation, ontology)
    pd.testing.assert_frame_equal(out, expected, check_like=True)


def test__mirror_ontology(annotation, ontology):
    aexp = pd.DataFrame(dict(mni_x=[-10, 30, 0],
                             structure_acronym=['S', 'Cl', 'CC'],
                             structure_id=[4260, 4322, 9422],
                             structure_name=['subiculum, right',
                                             'claustrum, left',
                                             'central canal']),
                        index=pd.Series(range(3), name='sample_id'))

    # this function doesn't touch mni_x -- it just assumes that all the
    # hemisphere designation are incorrect and updates the structure id, name,
    # and acronyms accordingly
    a = samples_._mirror_ontology(annotation, ontology)
    pd.testing.assert_frame_equal(a, aexp, check_like=True)


def test_mirror_samples(annotation, ontology):
    # we're changing quite a bit of stuff in the annotation dataframe
    aexp = pd.DataFrame(dict(mni_x=[-10, 30, 0, 10, -30],
                             structure_acronym=['S', 'Cl', 'CC', 'S', 'Cl'],
                             structure_id=[4251, 4323, 9422, 4260, 4322],
                             structure_name=['subiculum, left',
                                             'claustrum, right',
                                             'central canal',
                                             'subiculum, right',
                                             'claustrum, left']),
                        index=pd.Series([0, 1, 2, 0, 1], name='sample_id'))

    # but let's confirm all the outputs are as-expected
    a = samples_.mirror_samples(annotation, ontology)
    pd.testing.assert_frame_equal(a, aexp, check_like=True)


@pytest.mark.xfail
def test__assign_sample(atlas):
    atlas = atlas['image']
    assert samples_._assign_sample([[0, 0, 0]], atlas, tolerance=0) == 0
    assert samples_._assign_sample([[26, 96, 66]], atlas, tolerance=0) == 71
    assert False


@pytest.mark.xfail
def test__check_label():
    assert False


@pytest.mark.xfail
def test_label_samples():
    assert False


def test_groupby_index():
    # default usage (no params)
    microarray = pd.DataFrame([[0., 1.], [1., 2.], [5., 6.], [0., 1.]],
                              index=pd.Series([1, 1, 1, 2], name='label'))
    expected = pd.DataFrame([[2., 3.], [0., 1.]],
                            index=pd.Series([1, 2], name='label'))
    out = samples_.groupby_index(microarray)
    pd.testing.assert_frame_equal(out, expected, check_like=True)

    # supplying `labels` appends NaN rows to output
    expected = pd.DataFrame([[2., 3.], [0., 1.], [np.nan, np.nan]],
                            index=pd.Series([1, 2, 3], name='label'))
    out = samples_.groupby_index(microarray, labels=[1, 2, 3])
    pd.testing.assert_frame_equal(out, expected, check_like=True)

    # changing `metric` works
    expected = pd.DataFrame([[1., 2.], [0., 1.]],
                            index=pd.Series([1, 2], name='label'))
    out = samples_.groupby_index(microarray, metric='median')
    pd.testing.assert_frame_equal(out, expected, check_like=True)


def test_aggregate_samples():
    m1 = pd.DataFrame([[0., 1.], [1., 2.], [5., 6.], [0., 1.]],
                      index=pd.Series([1, 1, 1, 2], name='label'))
    m2 = pd.DataFrame([[0., 1.], [1., 2.], [5., 6.], [10., 11.]],
                      index=pd.Series([1, 1, 2, 3], name='label'))

    # region_agg='donors'
    expected = pd.DataFrame([[1.25, 2.25],
                             [2.5, 3.5],
                             [10., 11.],
                             [np.nan, np.nan]],
                            index=pd.Series([1, 2, 3, 4], name='label'))
    out = samples_.aggregate_samples([m1, m2], labels=[1, 2, 3, 4],
                                     region_agg='donors', agg_metric='mean')
    pd.testing.assert_frame_equal(out, expected, check_like=True)

    # region_agg = 'samples'
    expected = pd.DataFrame([[1.4, 2.4],
                             [2.5, 3.5],
                             [10., 11.],
                             [np.nan, np.nan]],
                            index=pd.Series([1, 2, 3, 4], name='label'))
    out = samples_.aggregate_samples([m1, m2], labels=[1, 2, 3, 4],
                                     region_agg='samples', agg_metric='mean')
    pd.testing.assert_frame_equal(out, expected, check_like=True)

    # return_donors=True, agg_metric='median'
    expected = [
        pd.DataFrame([[1., 2.], [0., 1.], [np.nan, np.nan], [np.nan, np.nan]],
                     index=pd.Series([1, 2, 3, 4], name='label')),
        pd.DataFrame([[0.5, 1.5], [5, 6], [10, 11], [np.nan, np.nan]],
                     index=pd.Series([1, 2, 3, 4], name='label'))
    ]
    out = samples_.aggregate_samples([m1, m2], labels=[1, 2, 3, 4],
                                     return_donors=True, agg_metric='median')
    for e, o in zip(expected, out):
        pd.testing.assert_frame_equal(o, e, check_like=True)

    # check that poorly normalized genes are removed
    m1.loc[2, 1], m2.loc[2, 1] = np.nan, np.nan
    expected = pd.DataFrame([[1.25],
                             [2.5],
                             [10],
                             [np.nan]],
                            index=pd.Series([1, 2, 3, 4], name='label'))
    out = samples_.aggregate_samples([m1, m2], labels=[1, 2, 3, 4],
                                     region_agg='donors', agg_metric='mean')
    pd.testing.assert_frame_equal(out, expected, check_like=True)

    # invalid method for region_agg
    with pytest.raises(ValueError):
        samples_.aggregate_samples([m1, m2], region_agg='notamethod')
    # region_agg='samples' incompatible with return_donors=True
    with pytest.raises(ValueError):
        samples_.aggregate_samples([m1, m2], region_agg='samples',
                                   return_donors=True)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# test all the above functions again on real data to make sure we don't choke #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


def test_update_mni_coords_real(testfiles):
    for annotation in flatten_dict(testfiles, 'annotation').values():
        samples_.update_mni_coords(annotation)


def test_label_samples_real(testfiles, atlas):
    out = samples_.label_samples(first_entry(testfiles, 'annotation'),
                                 atlas['image'])
    assert isinstance(out, pd.DataFrame)
    assert out.index.name == 'sample_id'
    assert out.columns == ['label']


def test_drop_mismatch_samples_real(testfiles):
    annotation = flatten_dict(testfiles, 'annotation').values()
    ontology = flatten_dict(testfiles, 'ontology').values()
    for an, on in zip(annotation, ontology):
        samples_.drop_mismatch_samples(an, on)


def test_mirror_samples_real(testfiles):
    annotation = flatten_dict(testfiles, 'annotation').values()
    ontology = flatten_dict(testfiles, 'ontology').values()
    orig = [363, 470]
    for an, on, o in zip(annotation, ontology, orig):
        out = samples_.mirror_samples(an, on)
        # there should be more than the original # of samples but less than or
        # equal to 2x that number (we can't MORE than duplicate)
        assert len(out) > o and len(out) <= o * 2


def test__mirror_ontology_real(testfiles):
    annotation = flatten_dict(testfiles, 'annotation').values()
    ontology = flatten_dict(testfiles, 'ontology').values()
    orig = [363, 470]
    for a, o, l in zip(annotation, ontology, orig):
        annot = samples_._mirror_ontology(a, o)
        assert len(annot) == l



"""
Functions for fetching and downloading files

Taken directly from `nilearn <https://github.com/nilearn/nilearn>`_ with some
light modifications. Nilearn is licensed under the BSD-3, duplicated here:

Copyright (c) 2007 - 2015 The nilearn developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

a. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
b. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
c. Neither the name of the nilearn developers nor the names of
    its contributors may be used to endorse or promote products
    derived from this software without specific prior written
    permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""

import base64
import contextlib
import hashlib
import os
import pickle
import shutil
import sys
import tarfile
import time
import urllib
import warnings
import zipfile


def _format_time(t):
    """ Pretty formatting of time `t`
    """
    if t > 60:
        return "%4.1fmin" % (t / 60.)
    else:
        return " %5.1fs" % (t)


def _md5_sum_file(path):
    """ Calculates the MD5 sum of a file.
    """
    with open(path, 'rb') as f:
        m = hashlib.md5()
        while True:
            data = f.read(8192)
            if not data:
                break
            m.update(data)
    return m.hexdigest()


def md5_hash(string):
    """ Calculates the MD5 sum of a string
    """
    m = hashlib.md5()
    m.update(string.encode('utf-8'))
    return m.hexdigest()


def movetree(src, dst):
    """ Move an entire tree to another directory, overwriting existing files
    """
    names = os.listdir(src)

    # Create destination dir if it does not exist
    if not os.path.exists(dst):
        os.makedirs(dst)
    errors = []

    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if os.path.isdir(srcname) and os.path.isdir(dstname):
                movetree(srcname, dstname)
                os.rmdir(srcname)
            else:
                shutil.move(srcname, dstname)
        except (IOError, os.error) as why:
            errors.append((srcname, dstname, str(why)))
        # catch the Error from the recursive movetree so that we can
        # continue with other files
        except Exception as err:
            errors.extend(err.args[0])
    if errors:
        raise Exception(errors)


def _chunk_report_(bytes_so_far, total_size, initial_size, t0):
    """Show downloading percentage.

    Parameters
    ----------
    bytes_so_far: int
        Number of downloaded bytes
    total_size: int
        Total size of the file (may be 0/None, depending on download method).
    t0: int
        The time in seconds (as returned by time.time()) at which the
        download was resumed / started.
    initial_size: int
        If resuming, indicate the initial size of the file.
        If not resuming, set to zero.
    """

    if not total_size:
        sys.stderr.write("\rDownloaded %d of ? bytes." % (bytes_so_far))

    else:
        # Estimate remaining download time
        total_percent = float(bytes_so_far) / total_size

        current_download_size = bytes_so_far - initial_size
        bytes_remaining = total_size - bytes_so_far
        dt = time.time() - t0
        download_rate = current_download_size / max(1e-8, float(dt))
        # Minimum rate of 0.01 bytes/s, to avoid dividing by zero.
        time_remaining = bytes_remaining / max(0.01, download_rate)

        # Trailing whitespace is to erase extra char when message length
        # varies
        sys.stderr.write(
            "\rDownloaded %d of %d bytes (%.1f%%, %s remaining)"
            % (bytes_so_far, total_size, total_percent * 100,
               _format_time(time_remaining)))


def _chunk_read_(response, local_file, chunk_size=8192, report_hook=None,
                 initial_size=0, total_size=None, verbose=1):
    """Download a file chunk by chunk and show advancement

    Parameters
    ----------
    response: _urllib.response.addinfourl
        Response to the download request in order to get file size
    local_file: file
        Hard disk file where data should be written
    chunk_size: int, optional
        Size of downloaded chunks. Default: 8192
    report_hook: bool
        Whether or not to show downloading advancement. Default: None
    initial_size: int, optional
        If resuming, indicate the initial size of the file
    total_size: int, optional
        Expected final size of download (None means it is unknown).
    verbose: int, optional
        verbosity level (0 means no message).

    Returns
    -------
    data: string
        The downloaded file.
    """
    try:
        if total_size is None:
            total_size = response.info().get('Content-Length').strip()
        total_size = int(total_size) + initial_size
    except Exception as e:
        if verbose > 2:
            print("Warning: total size could not be determined.")
            if verbose > 3:
                print("Full stack trace: %s" % e)
        total_size = None
    bytes_so_far = initial_size

    t0 = time_last_display = time.time()
    while True:
        chunk = response.read(chunk_size)
        bytes_so_far += len(chunk)
        time_last_read = time.time()
        if (report_hook
                # Refresh report every second or when download is
                # finished.
                and (time_last_read > time_last_display + 1. or not chunk)):
            _chunk_report_(bytes_so_far,
                           total_size, initial_size, t0)
            time_last_display = time_last_read
        if chunk:
            local_file.write(chunk)
        else:
            break

    return


def _fetch_file(url, data_dir, resume=True, overwrite=False,
                md5sum=None, username=None, password=None, handlers=[],
                verbose=1):
    """Load requested file, downloading it if needed or requested

    Parameters
    ----------
    url: string
        Contains the url of the file to be downloaded.
    data_dir: string
        Path of the data directory. Used for data storage in the specified
        location.
    resume: bool, optional
        If true, try to resume partially downloaded files
    overwrite: bool, optional
        If true and file already exists, delete it.
    md5sum: string, optional
        MD5 sum of the file. Checked if download of the file is required
    username: string, optional
        Username used for basic HTTP authentication
    password: string, optional
        Password used for basic HTTP authentication
    handlers: list of BaseHandler, optional
        urllib handlers passed to urllib.request.build_opener. Used by
        advanced users to customize request handling.
    verbose: int, optional
        verbosity level (0 means no message).

    Returns
    -------
    files: string
        Absolute path of downloaded file

    Notes
    -----
    If the download procedure fails all downloaded files are removed
    """

    # Determine data path
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    # Determine filename using URL
    parse = urllib.parse.urlparse(url)
    file_name = os.path.basename(parse.path)
    if file_name == '':
        file_name = md5_hash(parse.path)

    temp_file_name = file_name + ".part"
    full_name = os.path.join(data_dir, file_name)
    temp_full_name = os.path.join(data_dir, temp_file_name)
    if os.path.exists(full_name):
        if overwrite:
            os.remove(full_name)
        else:
            return full_name
    if os.path.exists(temp_full_name):
        if overwrite:
            os.remove(temp_full_name)
    t0 = time.time()
    local_file = None
    initial_size = 0

    try:
        # Download data
        url_opener = urllib.request.build_opener(*handlers)
        request = urllib.request.Request(url)
        request.add_header('Connection', 'Keep-Alive')
        if username is not None and password is not None:
            if not url.startswith('https'):
                raise ValueError(
                    'Authentication was requested on a non  secured URL (%s).'
                    'Request has been blocked for security reasons.' % url)
            # Note: HTTPBasicAuthHandler is not fitted here because it relies
            # on the fact that the server will return a 401 error with proper
            # www-authentication header, which is not the case of most
            # servers.
            encoded_auth = base64.b64encode(
                (username + ':' + password).encode())
            request.add_header(b'Authorization', b'Basic ' + encoded_auth)
        if verbose > 0:
            displayed_url = url.split('?')[0] if verbose == 1 else url
            print('Downloading data from %s ...' % displayed_url)
        if resume and os.path.exists(temp_full_name):
            # Download has been interrupted, we try to resume it.
            local_file_size = os.path.getsize(temp_full_name)
            # If the file exists, then only download the remainder
            request.add_header("Range", "bytes=%s-" % (local_file_size))
            try:
                data = url_opener.open(request)
                content_range = data.info().get('Content-Range')
                if (content_range is None or not content_range.startswith(
                        'bytes %s-' % local_file_size)):
                    raise IOError('Server does not support resuming')
            except Exception:
                # A wide number of errors can be raised here. HTTPError,
                # URLError... I prefer to catch them all and rerun without
                # resuming.
                if verbose > 0:
                    print('Resuming failed, try to download the whole file.')
                return _fetch_file(
                    url, data_dir, resume=False, overwrite=overwrite,
                    md5sum=md5sum, username=username, password=password,
                    handlers=handlers, verbose=verbose)
            local_file = open(temp_full_name, "ab")
            initial_size = local_file_size
        else:
            data = url_opener.open(request)
            local_file = open(temp_full_name, "wb")
        _chunk_read_(data, local_file, report_hook=(verbose > 0),
                     initial_size=initial_size, verbose=verbose)
        # temp file must be closed prior to the move
        if not local_file.closed:
            local_file.close()
        shutil.move(temp_full_name, full_name)
        dt = time.time() - t0
        if verbose > 0:
            # Complete the reporting hook
            sys.stderr.write(' ...done. ({0:.0f} seconds, {1:.0f} min)\n'
                             .format(dt, dt // 60))
    except (urllib.error.HTTPError, urllib.error.URLError):
        sys.stderr.write("Error while fetching file %s; dataset "
                         "fetching aborted." % (file_name))
        raise
    finally:
        if local_file is not None:
            if not local_file.closed:
                local_file.close()
    if md5sum is not None:
        if (_md5_sum_file(full_name) != md5sum):
            raise ValueError("File %s checksum verification has failed."
                             " Dataset fetching aborted." % local_file)
    return full_name


def _uncompress_file(file_, delete_archive=True, verbose=1):
    """Uncompress files contained in a data_set

    Parameters
    ----------
    file: string
        path of file to be uncompressed.
    delete_archive: bool, optional
        Wheteher or not to delete archive once it is uncompressed.
        Default: True
    verbose: int, optional
        verbosity level (0 means no message)

    Notes
    -----
    This handles zip, tar, gzip and bzip files only.
    """

    if verbose > 0:
        sys.stderr.write('Extracting data from %s...' % file_)
    data_dir = os.path.dirname(file_)
    # We first try to see if it is a zip file
    try:
        filename, ext = os.path.splitext(file_)
        with open(file_, "rb") as fd:
            header = fd.read(4)
        processed = False
        if zipfile.is_zipfile(file_):
            z = zipfile.ZipFile(file_)
            z.extractall(path=data_dir)
            z.close()
            if delete_archive:
                os.remove(file_)
            file_ = filename
            processed = True
        elif ext == '.gz' or header.startswith(b'\x1f\x8b'):
            import gzip
            gz = gzip.open(file_)
            if ext == '.tgz':
                filename = filename + '.tar'
            out = open(filename, 'wb')
            shutil.copyfileobj(gz, out, 8192)
            gz.close()
            out.close()
            # If file is .tar.gz, this will be handle in the next case
            if delete_archive:
                os.remove(file_)
            file_ = filename
            processed = True
        if os.path.isfile(file_) and tarfile.is_tarfile(file_):
            with contextlib.closing(tarfile.open(file_, "r")) as tar:
                tar.extractall(path=data_dir)
            if delete_archive:
                os.remove(file_)
            processed = True
        if not processed:
            raise IOError('[Uncompress] unknown archive file format: {}'
                          .format(file_))

        if verbose > 0:
            sys.stderr.write('.. done.\n')
    except Exception as e:
        if verbose > 0:
            print('Error uncompressing file: %s' % e)
        raise


def _fetch_files(data_dir, files, resume=True, mock=False, verbose=1):
    """Load requested dataset, downloading it if needed or requested

    This function retrieves files from the hard drive or download them from
    the given urls. Note to developpers: All the files will be first
    downloaded in a sandbox and, if everything goes well, they will be moved
    into the folder of the dataset. This prevents corrupting previously
    downloaded data. In case of a big dataset, do not hesitate to make several
    calls if needed.

    Parameters
    ----------
    data_dir: string
        Path of the data directory. Used for data storage in a specified
        location.
    files: list of (string, string, dict)
        List of files and their corresponding url with dictionary that contains
        options regarding the files. Eg. (file_path, url, opt). If a file_path
        is not found in data_dir, as in data_dir/file_path the download will
        be immediately cancelled and any downloaded files will be deleted.
        Options supported are:
            * 'move' if renaming the file or moving it to a subfolder is needed
            * 'uncompress' to indicate that the file is an archive
            * 'md5sum' to check the md5 sum of the file
            * 'overwrite' if the file should be re-downloaded even if it exists
    resume: bool, optional
        If true, try resuming download if possible
    mock: boolean, optional
        If true, create empty files if the file cannot be downloaded. Test use
        only.
    verbose: int, optional
        verbosity level (0 means no message).

    Returns
    -------
    files: list of string
        Absolute paths of downloaded files on disk
    """

    # There are two working directories here:
    # - data_dir is the destination directory of the dataset
    # - temp_dir is a temporary directory dedicated to this fetching call. All
    #   files that must be downloaded will be in this directory. If a corrupted
    #   file is found, or a file is missing, this working directory will be
    #   deleted.
    files = list(files)
    files_pickle = pickle.dumps([(file_, url) for file_, url, _ in files])
    files_md5 = hashlib.md5(files_pickle).hexdigest()
    temp_dir = os.path.join(data_dir, files_md5)

    # Create destination dir
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    # Abortion flag, in case of error
    abort = None

    files_ = []
    for file_, url, opts in files:
        # 3 possibilities:
        # - the file exists in data_dir, nothing to do.
        # - the file does not exists: we download it in temp_dir
        # - the file exists in temp_dir: this can happen if an archive has been
        #   downloaded. There is nothing to do

        # Target file in the data_dir
        target_file = os.path.join(data_dir, file_)
        # Target file in temp dir
        temp_target_file = os.path.join(temp_dir, file_)
        # Whether to keep existing files
        overwrite = opts.get('overwrite', False)
        if (abort is None and (overwrite
                               or (not os.path.exists(target_file)
                                   and not os.path.exists(temp_target_file)))):

            # We may be in a global read-only repository. If so, we cannot
            # download files.
            if not os.access(data_dir, os.W_OK):
                raise ValueError('Dataset files are missing but dataset'
                                 ' repository is read-only. Contact your data'
                                 ' administrator to solve the problem')

            if not os.path.exists(temp_dir):
                os.mkdir(temp_dir)
            md5sum = opts.get('md5sum', None)

            dl_file = _fetch_file(url, temp_dir, resume=resume,
                                  verbose=verbose, md5sum=md5sum,
                                  username=opts.get('username', None),
                                  password=opts.get('password', None),
                                  handlers=opts.get('handlers', []),
                                  overwrite=overwrite)
            if 'move' in opts:
                # XXX: here, move is supposed to be a dir, it can be a name
                move = os.path.join(temp_dir, opts['move'])
                move_dir = os.path.dirname(move)
                if not os.path.exists(move_dir):
                    os.makedirs(move_dir)
                shutil.move(dl_file, move)
                dl_file = move
            if 'uncompress' in opts:
                try:
                    if not mock or os.path.getsize(dl_file) != 0:
                        _uncompress_file(dl_file, verbose=verbose)
                    else:
                        os.remove(dl_file)
                except Exception as e:
                    abort = str(e)

        if (abort is None and not os.path.exists(target_file) and not
                os.path.exists(temp_target_file)):
            if not mock:
                warnings.warn('An error occured while fetching %s' % file_)
                abort = ("Dataset has been downloaded but requested file was "
                         "not provided:\nURL: %s\n"
                         "Target file: %s\nDownloaded: %s" %
                         (url, target_file, dl_file))
            else:
                if not os.path.exists(os.path.dirname(temp_target_file)):
                    os.makedirs(os.path.dirname(temp_target_file))
                open(temp_target_file, 'w').close()
        if abort is not None:
            if os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)
            raise IOError('Fetching aborted: ' + abort)
        files_.append(target_file)
    # If needed, move files from temps directory to final directory.
    if os.path.exists(temp_dir):
        # XXX We could only moved the files requested
        # XXX Movetree can go wrong
        movetree(temp_dir, data_dir)
        shutil.rmtree(temp_dir)
    return files_


def _get_dataset_dir(dataset_name, data_dir=None, verbose=1):
    """
    Gets path to `dataset_name`

    Parameters
    ----------
    dataset_name : str
        The name of the dataset in question
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variables 'ABAGEN_DATA'; if that is not set, will use
        '~/abagen-data' instead. Default: None
    verbose : int, optional
        Verbosity level (0 means no message). Default: 1

    Returns
    -------
    data_dir : str
        Path to use as data directory
    """

    paths = [
        os.environ.get('ABAGEN_DATA'),
        os.path.join('~', 'abagen-data'),
        os.getcwd()
    ]
    if data_dir is not None:
        paths = [data_dir, os.path.dirname(data_dir)] + paths
    paths = [os.path.expanduser(d) for d in paths if d is not None]

    if verbose > 2:
        print('Dataset search paths: {}'.format(paths))

    for path in paths:
        if not os.path.basename(path) == dataset_name:
            path = os.path.join(path, dataset_name)
        if os.path.islink(path):
            link = os.readlink(path)
            if os.path.isabs(link):
                path = link
            path = os.path.join(os.path.dirname(path), link)
        if os.path.exists(path) and os.path.isdir(path):
            if verbose > 1:
                print('\nDataset found in {}\n'.format(path))
            return path

    errors = []
    for path in paths:
        if not os.path.basename(path) == dataset_name:
            path = os.path.join(path, dataset_name)
        if not os.path.exists(path):
            try:
                os.makedirs(path)
                if verbose > 0:
                    print('\nDataset created in {}\n'.format(path))
                return path
            except (NotADirectoryError, PermissionError) as exc:
                err_msg = getattr(exc, 'strerror', str(exc))
                errors.append('\n -{} ({})'.format(path, err_msg))

    raise OSError('Tried to store dataset {} in the following directories, '
                  'but: ' + ''.join(errors))



# -*- coding: utf-8 -*-

import numpy as np
import pytest
from abagen.mouse import mouse

STRUCTURES = [182305713, 182305709, 182305705]
ATTRIBUTES = ['expression_energy', 'expression_density', 'sum_pixels']
EXPERIMENTS = {
    986: {
        'expression_energy': np.array([7.73432, 7.28206, 3.82741]),
        'expression_density': np.array([0.0603072, 0.0553335, 0.0298628]),
        'sum_pixels': np.array([419628.0, 2238000.0, 3629050.0])
    },
    69782969: {
        'expression-energy': np.array([1.00879, 1.40198, 2.34988]),
        'expression-density': np.array([0.00650952, 0.00896709, 0.015409]),
        'sum-pixels': np.array([195863.0, 1051230.0, 1440430.0])
    }
}
GBA_UNIONIZATION = {
    'expression_energy': np.array([0.236301, 0.261266, 0.416281]),
    'expression_density': np.array([0.00207073, 0.00232453, 0.00372069]),
    'sum_pixels': np.array([419628.0, 2238000.0, 3636670.0]),
    'voxel_energy_cv': np.array([0.923913, 0.995631, 0.81265])
}


@pytest.mark.parametrize(('genes', 'direction', 'expected'), [
    (dict(acronym='Snca'), 'sagittal', [988, 990, 79904550]),
    (dict(acronym='Snca'), 'coronal', [986, 79908848]),
    (dict(acronym='Gba'), 'sagittal', [1612]),
    (dict(acronym='Gba'), 'coronal', [1611]),
    (dict(acronym='Elf4'), 'sagittal', [73834415, 77464840]),
    (dict(id=84193), 'sagittal', [70238925, 71147924, 71213117]),
    (dict(id=84193), 'coronal', [74047443]),
    (dict(id=18608), 'sagittal', [69289721]),
])
def test_get_experiments_from_gene(genes, direction, expected):
    with pytest.raises(ValueError):
        mouse._get_experiments_from_gene(acronym='notagene',
                                         slicing_direction='coronal')
    with pytest.raises(ValueError):
        mouse._get_experiments_from_gene(id=84193,
                                         slicing_direction='notadirection')
    with pytest.raises(ValueError):
        mouse._get_experiments_from_gene(id=-1000000)

    # get experiments from provided input
    exp = mouse._get_experiments_from_gene(**genes,
                                           slicing_direction=direction)
    assert len(exp) == len(expected)
    assert all(i == j for (i, j) in zip(sorted(exp), sorted(expected)))


@pytest.mark.parametrize(('experiment', 'attributes'), [
    (986, None), (986, ATTRIBUTES), (986, 'all'),
    (69782969, None), (69782969, ATTRIBUTES), (69782969, 'all')
])
def test_get_unionization_from_experiment(experiment, attributes):
    with pytest.raises(ValueError):
        mouse._get_unionization_from_experiment(-100, structures=STRUCTURES)

    # get data from provided experiment
    data = mouse._get_unionization_from_experiment(experiment,
                                                   structures=STRUCTURES,
                                                   attributes=attributes)
    data.index = data.index.droplevel('gene_id')

    if attributes is None:
        attributes = ['expression_density']
    elif attributes == 'all':
        attributes = mouse._UNIONIZATION_ATTRIBUTES

    assert len(data.columns) == len(attributes)
    for attr in set(EXPERIMENTS[experiment].keys()).intersection(attributes):
        assert np.allclose(np.asarray(data.loc[STRUCTURES, attr]),
                           EXPERIMENTS[experiment][attr])


@pytest.mark.parametrize(('attribute'), [
    'expression_energy', 'expression_density', 'sum_pixels', 'voxel_energy_cv'
])
def test_get_unionization_from_gene(attribute):
    with pytest.raises(ValueError):
        mouse.get_unionization_from_gene(acronym='Gba',
                                         slicing_direction='notadirection',
                                         structures=STRUCTURES)
    with pytest.raises(ValueError):
        mouse.get_unionization_from_gene(id=18608,
                                         slicing_direction='coronal',
                                         structures=STRUCTURES)

    # get data for provided gene
    data = mouse.get_unionization_from_gene(acronym='Gba',
                                            slicing_direction='coronal',
                                            structures=STRUCTURES,
                                            attributes=attribute)
    data.index = data.index.droplevel('gene_id')

    assert np.allclose(data.loc[STRUCTURES, attribute],
                       GBA_UNIONIZATION[attribute])
